import numpy as np
from typing import Optional

from arcticpy import add_cti
from arcticpy import remove_cti

from arcticpy import ROE

import autoarray as aa

from autocti.clocker.abstract import AbstractClocker
from autocti.model.model_util import CTI2D
from autocti.preloads import Preloads


class Clocker2D(AbstractClocker):
    def __init__(
        self,
        iterations: int = 5,
        parallel_roe: Optional[ROE] = None,
        parallel_express: int = 0,
        parallel_window_offset: int = 0,
        parallel_window_start: int = 0,
        parallel_window_stop: int = -1,
        parallel_time_start=0,
        parallel_time_stop=-1,
        parallel_prune_n_electrons=1e-18,
        parallel_prune_frequency=0,
        parallel_poisson_traps: bool = False,
        parallel_fast_mode: Optional[bool] = False,
        serial_roe: Optional[ROE] = None,
        serial_express: int = 0,
        serial_window_offset: int = 0,
        serial_window_start: int = 0,
        serial_window_stop: int = -1,
        serial_time_start=0,
        serial_time_stop=-1,
        serial_prune_n_electrons=1e-18,
        serial_prune_frequency=0,
        serial_fast_mode: Optional[bool] = None,
        allow_negative_pixels=1,
        pixel_bounce=None,
        verbosity: int = 0,
        poisson_seed: int = -1,
    ):
        """
        Performs clocking of a 2D image via the c++ arctic algorithm.

        This corresponds to a full read out of a CCD including clocking in the parallel and / or serial direction.

        Parameters
        ----------
        iterations
            The number of iterations used to correct CTI from an image.
        parallel_roe
            Contains parameters describing the read-out electronics of the CCD (e.g. CCD dwell times, charge injection
            clocking, etc.) for clocking in the parallel direction.
        parallel_express
            An integer factor describing how parallel pixel-to-pixel transfers are combined into single transfers for
            efficiency (see: https://academic.oup.com/mnras/article/401/1/371/1006825).
        parallel_window_offset
            The number of pixels before parallel clocking begins, thereby extending the length over which clocking
            is performed in the parallel direction and increasing CTI.
        parallel_window_start
            The pixel index of the input image where parallel arCTIc clocking begins, for example
            if `window_start=10` the first 10 pixels are omitted and not clocked.
        parallel_window_start
            The pixel index of the input image where parallel arCTIc clocking ends, for example if `window_start=20`
            any pixels after the 20th pixel are omitted and not clocked.
        parallel_poisson_traps
            If `True`, the density of every trap species (which in the `Trap` objects is defined as the mean density
            of that species) are drawn from a Poisson distribution for every column of data.
        parallel_fast_mode
            If input, parallel CTI is added via arctic efficiently by determining all unique columns in the input data,
            calling arctic once per unique column and mapping the 1D output over the full 2D image. This requires
            many column in the image to have the same signal to be efficient, for example charge injection
            calibration data.
        serial_roe
            Contains parameters describing the read-out electronics of the CCD (e.g. CCD dwell times, charge injection
            clocking, etc.) for clocking in the serial direction.
        serial_express
            An integer factor describing how serial pixel-to-pixel transfers are combined into single transfers for
            efficiency (see: https://academic.oup.com/mnras/article/401/1/371/1006825).
        serial_window_offset
            The number of pixels before serial clocking begins, thereby extending the length over which clocking
            is performed in the serial direction and increasing CTI.
        serial_window_start
            The pixel index of the input image where serial arCTIc clocking begins, for example
            if `window_start=10` the first 10 pixels are omitted and not clocked.
        serial_window_start
            The pixel index of the input image where serial arCTIc clocking ends, for example if `window_start=20`
            any pixels after the 20th pixel are omitted and not clocked.
        serial_fast_mode
            If input, serial CTI is added via arctic efficiently by calling arctic once and mapping the 1D output over
            the full 2D image. This requires every row in the image has the same signal (such that each column gives
            an identical arctic output).
        allow_negative_pixels
            If True, negative electrons in a pixel are allowed and modeled via arCTIc, if Falss they are explicitly
            not allowed.
        verbosity
            Whether to silence print statements and output from the c++ arctic call.
        poisson_seed
            A seed for the random number generator which draws the Poisson trap densities from a Poisson distribution.
        """

        super().__init__(iterations=iterations, verbosity=verbosity)

        self.parallel_roe = parallel_roe or ROE()
        self.parallel_express = parallel_express
        self.parallel_window_offset = parallel_window_offset
        self.parallel_window_start = parallel_window_start
        self.parallel_window_stop = parallel_window_stop
        self.parallel_time_start = parallel_time_start
        self.parallel_time_stop = parallel_time_stop
        self.parallel_prune_n_electrons = parallel_prune_n_electrons
        self.parallel_prune_frequency = parallel_prune_frequency
        self.parallel_poisson_traps = parallel_poisson_traps
        self.parallel_fast_mode = parallel_fast_mode

        self.serial_roe = serial_roe or ROE()
        self.serial_express = serial_express
        self.serial_window_offset = serial_window_offset
        self.serial_window_start = serial_window_start
        self.serial_window_stop = serial_window_stop
        self.serial_time_start = serial_time_start
        self.serial_time_stop = serial_time_stop
        self.serial_prune_n_electrons = serial_prune_n_electrons
        self.serial_prune_frequency = serial_prune_frequency
        self.serial_fast_mode = serial_fast_mode

        self.allow_negative_pixels = allow_negative_pixels
        self.pixel_bounce = pixel_bounce

        self.poisson_seed = poisson_seed

    def _parallel_traps_ccd_from(self, cti: CTI2D):
        """
        Unpack the `CTI1D` object to retries its traps and ccd.
        """
        if cti.parallel_trap_list is not None:
            trap_list = list(cti.parallel_trap_list)
        else:
            trap_list = None

        return trap_list, cti.parallel_ccd

    def _serial_traps_ccd_from(self, cti: CTI2D):
        """
        Unpack the `CTI1D` object to retries its traps and ccd.
        """
        if cti.serial_trap_list is not None:
            trap_list = list(cti.serial_trap_list)
        else:
            trap_list = None

        return trap_list, cti.serial_ccd

    def add_cti(
        self, data: aa.Array2D, cti: CTI2D, preloads: Optional[Preloads] = Preloads()
    ) -> aa.Array2D:
        """
        Add CTI to a 2D dataset by passing it to the c++ arctic clocking algorithm.

        Clocking is performed towards the readout register and electronics, with parallel CTI added first followed
        by serial CTI. If both parallel and serial CTI are added, parallel CTI is added and the post-CTI image
        (therefore including trailing after parallel clocking) is used to perform serial clocking and add serial CTI.

        If the flag `parallel_poisson_traps=True` the parallel clocking algorithm is sent through the alternative
        function `add_cti_poisson_traps`.  This adds CTI via arctic in the same way, however for parallel clocking
        the density of traps in every column is drawn from a Poisson distribution to represent the stochastic
        nature of how many traps are in each column of a real CCD.

        Parameters
        ----------
        data
            The 1D data that is clocked via arctic and has CTI added to it.
        cti
            An object which represents the CTI properties of 2D clocking, including the trap species which capture
            and release electrons and the volume-filling behaviour of the CCD for parallel and serial clocking.
        """

        if self.parallel_poisson_traps:
            return self.add_cti_poisson_traps(data=data, cti=cti)

        if self.parallel_fast_mode:
            return self.add_cti_parallel_fast(data=data, cti=cti, preloads=preloads)

        if self.serial_fast_mode:
            return self.add_cti_serial_fast(data=data, cti=cti, preloads=preloads)

        data = data.native_skip_mask

        parallel_trap_list, parallel_ccd = self._parallel_traps_ccd_from(cti=cti)
        serial_trap_list, serial_ccd = self._serial_traps_ccd_from(cti=cti)

        self.check_traps(trap_list_0=parallel_trap_list, trap_list_1=serial_trap_list)
        self.check_ccd(ccd_list=[parallel_ccd, serial_ccd])

        parallel_ccd = self.ccd_from(ccd_phase=parallel_ccd)
        serial_ccd = self.ccd_from(ccd_phase=serial_ccd)

        try:
            parallel_window_offset = data.readout_offsets[0]
            serial_window_offset = data.readout_offsets[1]
        except AttributeError:
            parallel_window_offset = self.parallel_window_offset
            serial_window_offset = self.serial_window_offset

        try:
            image_post_cti = add_cti(
                image=data,
                parallel_ccd=parallel_ccd,
                parallel_roe=self.parallel_roe,
                parallel_traps=parallel_trap_list,
                parallel_express=self.parallel_express,
                parallel_window_offset=parallel_window_offset,
                parallel_window_start=self.parallel_window_start,
                parallel_window_stop=self.parallel_window_stop,
                parallel_time_start=self.parallel_time_start,
                parallel_time_stop=self.parallel_time_stop,
                parallel_prune_n_electrons=self.parallel_prune_n_electrons,
                parallel_prune_frequency=self.parallel_prune_frequency,
                serial_ccd=serial_ccd,
                serial_roe=self.serial_roe,
                serial_traps=serial_trap_list,
                serial_express=self.serial_express,
                serial_window_offset=serial_window_offset,
                serial_window_start=self.serial_window_start,
                serial_window_stop=self.serial_window_stop,
                serial_time_start=self.serial_time_start,
                serial_time_stop=self.serial_time_stop,
                serial_prune_n_electrons=self.serial_prune_n_electrons,
                serial_prune_frequency=self.serial_prune_frequency,
                allow_negative_pixels=self.allow_negative_pixels,
                pixel_bounce=self.pixel_bounce,
                verbosity=self.verbosity,
            )
        except TypeError:
            image_post_cti = add_cti(
                image=data,
                parallel_ccd=parallel_ccd,
                parallel_roe=self.parallel_roe,
                parallel_traps=parallel_trap_list,
                parallel_express=self.parallel_express,
                parallel_window_offset=parallel_window_offset,
                parallel_window_start=self.parallel_window_start,
                parallel_window_stop=self.parallel_window_stop,
                parallel_time_start=self.parallel_time_start,
                parallel_time_stop=self.parallel_time_stop,
                parallel_prune_n_electrons=self.parallel_prune_n_electrons,
                parallel_prune_frequency=self.parallel_prune_frequency,
                serial_ccd=serial_ccd,
                serial_roe=self.serial_roe,
                serial_traps=serial_trap_list,
                serial_express=self.serial_express,
                serial_window_offset=serial_window_offset,
                serial_window_start=self.serial_window_start,
                serial_window_stop=self.serial_window_stop,
                serial_time_start=self.serial_time_start,
                serial_time_stop=self.serial_time_stop,
                serial_prune_n_electrons=self.serial_prune_n_electrons,
                serial_prune_frequency=self.serial_prune_frequency,
                pixel_bounce=self.pixel_bounce,
                verbosity=self.verbosity,
            )

        try:
            return aa.Array2D(
                values=image_post_cti, mask=data.mask, store_native=True, skip_mask=True
            )
        except AttributeError:
            return image_post_cti

    def add_cti_poisson_traps(self, data: aa.Array2D, cti: CTI2D) -> aa.Array2D:
        """
        Add CTI to a 2D dataset by passing it to the c++ arctic clocking algorithm.

        Clocking is performed towards the readout register and electronics, with parallel CTI added first followed
        by serial CTI. If both parallel and serial CTI are added, parallel CTI is added and the post-CTI image
        (therefore including trailing after parallel clocking) is used to perform serial clocking and add serial CTI.

        This algorithm adds CTI via arctic in the same way as the normal `add_cti` function, however for parallel
        clocking the density of traps in every column is drawn from a Poisson distribution to represent the stochastic
        nature of how many traps are in each column of a real CCD.

        Parameters
        ----------
        data
            The 1D data that is clocked via arctic and has CTI added to it.
        cti
            An object which represents the CTI properties of 2D clocking, including the trap species which capture
            and release electrons and the volume-filling behaviour of the CCD for parallel and serial clocking.
        """

        parallel_trap_list, parallel_ccd = self._parallel_traps_ccd_from(cti=cti)
        serial_trap_list, serial_ccd = self._serial_traps_ccd_from(cti=cti)

        self.check_traps(trap_list_0=parallel_trap_list, trap_list_1=serial_trap_list)
        self.check_ccd(ccd_list=[parallel_ccd, serial_ccd])

        parallel_ccd = self.ccd_from(ccd_phase=parallel_ccd)

        try:
            parallel_window_offset = data.readout_offsets[0]
        except AttributeError:
            parallel_window_offset = self.parallel_window_offset

        image_pre_cti = data.native_skip_mask
        image_post_cti = np.zeros(data.shape_native)

        total_rows = image_post_cti.shape[0]
        total_columns = image_post_cti.shape[1]

        parallel_trap_column_list = []

        for column in range(total_columns):
            parallel_trap_poisson_list = [
                parallel_trap.poisson_density_from(
                    total_pixels=total_rows, seed=self.poisson_seed
                )
                for parallel_trap in parallel_trap_list
            ]

            parallel_trap_column_list.append(parallel_trap_poisson_list)

            image_pre_cti_pass = np.zeros(shape=(total_rows, 1))
            image_pre_cti_pass[:, 0] = image_pre_cti[:, column]

            try:
                image_post_cti[:, column] = add_cti(
                    image=image_pre_cti_pass,
                    parallel_ccd=parallel_ccd,
                    parallel_roe=self.parallel_roe,
                    parallel_traps=parallel_trap_poisson_list,
                    parallel_express=self.parallel_express,
                    parallel_window_offset=parallel_window_offset,
                    parallel_window_start=self.parallel_window_start,
                    parallel_window_stop=self.parallel_window_stop,
                    allow_negative_pixels=self.allow_negative_pixels,
                    pixel_bounce=self.pixel_bounce,
                    verbosity=self.verbosity,
                )[:, 0]
            except TypeError:
                image_post_cti[:, column] = add_cti(
                    image=image_pre_cti_pass,
                    parallel_ccd=parallel_ccd,
                    parallel_roe=self.parallel_roe,
                    parallel_traps=parallel_trap_poisson_list,
                    parallel_express=self.parallel_express,
                    parallel_window_offset=parallel_window_offset,
                    parallel_window_start=self.parallel_window_start,
                    parallel_window_stop=self.parallel_window_stop,
                    pixel_bounce=self.pixel_bounce,
                    verbosity=self.verbosity,
                )[:, 0]

        self.parallel_trap_column_list = parallel_trap_column_list

        serial_ccd = self.ccd_from(ccd_phase=serial_ccd)

        try:
            serial_window_offset = data.readout_offsets[1]
        except AttributeError:
            serial_window_offset = self.serial_window_offset

        try:
            image_post_cti = add_cti(
                image=image_post_cti,
                serial_ccd=serial_ccd,
                serial_roe=self.serial_roe,
                serial_traps=serial_trap_list,
                serial_express=self.serial_express,
                serial_window_offset=serial_window_offset,
                serial_window_start=self.serial_window_start,
                serial_window_stop=self.serial_window_stop,
                serial_time_start=self.serial_time_start,
                serial_time_stop=self.serial_time_stop,
                serial_prune_n_electrons=self.serial_prune_n_electrons,
                serial_prune_frequency=self.serial_prune_frequency,
                allow_negative_pixels=self.allow_negative_pixels,
                pixel_bounce=self.pixel_bounce,
                verbosity=self.verbosity,
            )
        except TypeError:
            image_post_cti = add_cti(
                image=image_post_cti,
                serial_ccd=serial_ccd,
                serial_roe=self.serial_roe,
                serial_traps=serial_trap_list,
                serial_express=self.serial_express,
                serial_window_offset=serial_window_offset,
                serial_window_start=self.serial_window_start,
                serial_window_stop=self.serial_window_stop,
                serial_time_start=self.serial_time_start,
                serial_time_stop=self.serial_time_stop,
                serial_prune_n_electrons=self.serial_prune_n_electrons,
                serial_prune_frequency=self.serial_prune_frequency,
                pixel_bounce=self.pixel_bounce,
                verbosity=self.verbosity,
            )

        try:
            return aa.Array2D(
                values=image_post_cti, mask=data.mask, store_native=True, skip_mask=True
            )
        except AttributeError:
            return image_post_cti

    def fast_indexes_from(self, data: aa.Array2D, for_parallel: bool):
        """
        For 2D images where the same signal is repeated over many columns (e.g. uniform charge injection imaging)
        the CTI added to each column for parallel CTI via arctic is identical. The same is true of serial CTI
        addition, where many rows of data are repeated.

        Therefore, this function speeds up CTI addition via arCTIc by extracting all identical columns (for parallel
        clocking) or rows (for serial clocking), adding CTI via arcitc to only these extracted stripes and copying the
        output stripes to construct the final post-cti image.

        This function inspects the input image and extracts the indexes of every unique stripe, alongside
        a list which maps this unique stripe to all other stripes it is identical to. These are used in the functions
        `add_cti_parallel_fast` and `add_cti_serial_fast` to create the extracted image that is passed to arctic and
        rebuild the final post CTI image.

        Parameters
        ----------
        data
            The 1D data that is clocked via arctic and has CTI added to it.
        """

        if for_parallel:
            total_stripes = data.shape[1]
        else:
            total_stripes = data.shape[0]

        fast_index_list = []
        fast_column_lists = []

        unchecked_list = range(0, total_stripes)

        for stripe_index in range(total_stripes):
            paired = False

            pair_list = []
            unchecked_list_new = []

            if stripe_index in unchecked_list:
                for pair_index in unchecked_list:
                    if for_parallel:
                        residual_map = np.abs(
                            data[:, stripe_index] - data[:, pair_index]
                        )
                    else:
                        residual_map = np.abs(
                            data[stripe_index, :] - data[pair_index, :]
                        )

                    if np.all(residual_map < 1.0e-8):
                        if not paired:
                            fast_index_list.append(stripe_index)

                        paired = True

                        pair_list.append(pair_index)

                    else:
                        unchecked_list_new.append(pair_index)

                fast_column_lists.append(pair_list)
                unchecked_list = unchecked_list_new

        return fast_index_list, fast_column_lists

    def add_cti_parallel_fast(
        self, data: aa.Array2D, cti: CTI2D, preloads: Preloads = Preloads()
    ):
        """
        Add CTI to a 2D dataset by passing it to the c++ arctic clocking algorithm.

        Clocking is performed towards the readout register and electronics, with parallel CTI added first followed
        by serial CTI. If both parallel and serial CTI are added, parallel CTI is added and the post-CTI image
        (therefore including trailing after parallel clocking) is used to perform serial clocking and add serial CTI.

        For 2D images where the same signal is repeated over many columns (e.g. uniform charge injection imaging)
        the CTI added to each column via arctic is identical. Therefore, this function speeds up CTI addition by
        extracting all identical columns, adding CTI via arcitc to only these columns and copying the output columns
        to construct the final post-cti image.

        Parameters
        ----------
        data
            The 1D data that is clocked via arctic and has CTI added to it.
        cti
            An object which represents the CTI properties of 2D clocking, including the trap species which capture
            and release electrons and the volume-filling behaviour of the CCD for parallel clocking.
        perform_checks
            Check that it is value for the input data to use the fast clocking speed up (e.g. all columns are identical
            and all entries outside this region are zero).
        """

        parallel_trap_list, parallel_ccd = self._parallel_traps_ccd_from(cti=cti)

        image_pre_cti = data.native_skip_mask

        self.check_traps(trap_list_0=parallel_trap_list)
        self.check_ccd(ccd_list=[parallel_ccd])

        parallel_ccd = self.ccd_from(ccd_phase=parallel_ccd)

        if preloads.parallel_fast_index_list is None:
            fast_index_list, fast_column_lists = self.fast_indexes_from(
                data=image_pre_cti, for_parallel=True
            )
        else:
            fast_index_list = preloads.parallel_fast_index_list
            fast_column_lists = preloads.parallel_fast_column_lists

        image_pre_cti_pass = np.zeros(
            shape=(data.shape_native[0], len(fast_index_list))
        )
        for i, fast_index in enumerate(fast_index_list):
            image_pre_cti_pass[:, i] = image_pre_cti[:, fast_index]

        try:
            image_post_cti_pass = add_cti(
                image=image_pre_cti_pass,
                parallel_ccd=parallel_ccd,
                parallel_roe=self.parallel_roe,
                parallel_traps=parallel_trap_list,
                parallel_express=self.parallel_express,
                parallel_window_offset=self.parallel_window_offset,
                parallel_time_start=self.parallel_time_start,
                parallel_time_stop=self.parallel_time_stop,
                parallel_prune_n_electrons=self.parallel_prune_n_electrons,
                parallel_prune_frequency=self.parallel_prune_frequency,
                allow_negative_pixels=self.allow_negative_pixels,
                pixel_bounce=self.pixel_bounce,
                verbosity=self.verbosity,
            )
        except TypeError:
            image_post_cti_pass = add_cti(
                image=image_pre_cti_pass,
                parallel_ccd=parallel_ccd,
                parallel_roe=self.parallel_roe,
                parallel_traps=parallel_trap_list,
                parallel_express=self.parallel_express,
                parallel_window_offset=self.parallel_window_offset,
                parallel_time_start=self.parallel_time_start,
                parallel_time_stop=self.parallel_time_stop,
                parallel_prune_n_electrons=self.parallel_prune_n_electrons,
                parallel_prune_frequency=self.parallel_prune_frequency,
                pixel_bounce=self.pixel_bounce,
                verbosity=self.verbosity,
            )

        image_post_cti = np.zeros(shape=data.shape_native)

        for i, fast_column_list in enumerate(fast_column_lists):
            for fast_column in fast_column_list:
                image_post_cti[:, fast_column] = image_post_cti_pass[:, i]

        if cti.serial_trap_list is None:
            return aa.Array2D(
                values=image_post_cti, mask=data.mask, store_native=True, skip_mask=True
            )

        serial_trap_list, serial_ccd = self._serial_traps_ccd_from(cti=cti)

        self.check_traps(trap_list_0=serial_trap_list)
        self.check_ccd(ccd_list=[serial_ccd])

        serial_ccd = self.ccd_from(ccd_phase=serial_ccd)

        try:
            image_post_cti = add_cti(
                image=image_post_cti,
                serial_ccd=serial_ccd,
                serial_roe=self.serial_roe,
                serial_traps=serial_trap_list,
                serial_express=self.serial_express,
                serial_window_offset=self.serial_window_offset,
                serial_time_start=self.serial_time_start,
                serial_time_stop=self.serial_time_stop,
                serial_prune_n_electrons=self.serial_prune_n_electrons,
                serial_prune_frequency=self.serial_prune_frequency,
                allow_negative_pixels=self.allow_negative_pixels,
                pixel_bounce=self.pixel_bounce,
                verbosity=self.verbosity,
            )
        except TypeError:
            image_post_cti = add_cti(
                image=image_post_cti,
                serial_ccd=serial_ccd,
                serial_roe=self.serial_roe,
                serial_traps=serial_trap_list,
                serial_express=self.serial_express,
                serial_window_offset=self.serial_window_offset,
                serial_time_start=self.serial_time_start,
                serial_time_stop=self.serial_time_stop,
                serial_prune_n_electrons=self.serial_prune_n_electrons,
                serial_prune_frequency=self.serial_prune_frequency,
                pixel_bounce=self.pixel_bounce,
                verbosity=self.verbosity,
            )

        return aa.Array2D(
            values=image_post_cti, mask=data.mask, store_native=True, skip_mask=True
        )

    def add_cti_serial_fast(
        self, data: aa.Array2D, cti: CTI2D, preloads: Preloads = Preloads()
    ):
        """
        Add CTI to a 2D dataset by passing it to the c++ arctic clocking algorithm.

        Clocking is performed towards the readout register and electronics, with parallel CTI added first followed
        by serial CTI. If both parallel and serial CTI are added, serial CTI is added and the post-CTI image
        (therefore including trailing after serial clocking) is used to perform serial clocking and add serial CTI.

        For 2D images where the same signal is repeated over all columns (e.g. uniform charge injection imaging)
        the CTI added to each row via arctic is identical. Therefore, this function speeds up CTI addition by
        only a single column to arcitc once and copying the output column the NumPy array to construct the final
        post-cti image.

        This only works for serial CTI when parallel CTI is omitted.

        By default, checks are performed which ensure that the input data fits the criteria for this speed up.

        Parameters
        ----------
        data
            The 1D data that is clocked via arctic and has CTI added to it.
        cti
            An object which represents the CTI properties of 2D clocking, including the trap species which capture
            and release electrons and the volume-filling behaviour of the CCD for serial clocking.
        perform_checks
            Check that it is value for the input data to use the fast clocking speed up (e.g. all columns are identical
            and all entries outside this region are zero).
        """

        serial_trap_list, serial_ccd = self._serial_traps_ccd_from(cti=cti)

        image_pre_cti = data.native_skip_mask

        self.check_traps(trap_list_0=serial_trap_list)
        self.check_ccd(ccd_list=[serial_ccd])

        serial_ccd = self.ccd_from(ccd_phase=serial_ccd)

        if preloads.serial_fast_index_list is None:
            fast_index_list, fast_row_lists = self.fast_indexes_from(
                data=image_pre_cti, for_parallel=False
            )
        else:
            fast_index_list = preloads.serial_fast_index_list
            fast_row_lists = preloads.serial_fast_row_lists

        image_pre_cti_pass = np.zeros(
            shape=(len(fast_index_list), data.shape_native[1])
        )
        for i, fast_index in enumerate(fast_index_list):
            image_pre_cti_pass[i, :] = image_pre_cti[fast_index, :]

        try:
            image_post_cti_pass = add_cti(
                image=image_pre_cti_pass,
                serial_ccd=serial_ccd,
                serial_roe=self.serial_roe,
                serial_traps=serial_trap_list,
                serial_express=self.serial_express,
                serial_window_offset=self.serial_window_offset,
                serial_time_start=self.serial_time_start,
                serial_time_stop=self.serial_time_stop,
                serial_prune_n_electrons=self.serial_prune_n_electrons,
                serial_prune_frequency=self.serial_prune_frequency,
                allow_negative_pixels=self.allow_negative_pixels,
                pixel_bounce=self.pixel_bounce,
                verbosity=self.verbosity,
            )
        except TypeError:
            image_post_cti_pass = add_cti(
                image=image_pre_cti_pass,
                serial_ccd=serial_ccd,
                serial_roe=self.serial_roe,
                serial_traps=serial_trap_list,
                serial_express=self.serial_express,
                serial_window_offset=self.serial_window_offset,
                serial_time_start=self.serial_time_start,
                serial_time_stop=self.serial_time_stop,
                serial_prune_n_electrons=self.serial_prune_n_electrons,
                serial_prune_frequency=self.serial_prune_frequency,
                pixel_bounce=self.pixel_bounce,
                verbosity=self.verbosity,
            )

        image_post_cti = np.zeros(shape=data.shape_native)

        for i, fast_row_list in enumerate(fast_row_lists):
            for fast_row in fast_row_list:
                image_post_cti[fast_row, :] = image_post_cti_pass[i, :]

        return aa.Array2D(
            values=image_post_cti, mask=data.mask, store_native=True, skip_mask=True
        )

    def remove_cti(self, data: aa.Array2D, cti: CTI2D) -> aa.Array2D:
        """
        Remove CTI from a 2D dataset by passing it to the c++ arctic clocking algorithm. The removal of CTI is
        performed by adding CTI to the data to understand how electrons are moved on the CCD, and using this
        image to then move them back to their original pixel.

        Clocking is performed towards the readout register and electronics, with parallel CTI added first followed
        by serial CTI. If both parallel and serial CTI are added, parallel CTI is added and the post-CTI image
        (therefore including trailing after parallel clocking) is used to perform serial clocking and add serial CTI.

        Parameters
        ----------
        data
            The 1D data that is clocked via arctic and has CTI added to it.
        cti
            An object which represents the CTI properties of 2D clocking, including the trap species which capture
            and release electrons and the volume-filling behaviour of the CCD for parallel and serial clocking.
        """

        parallel_trap_list, parallel_ccd = self._parallel_traps_ccd_from(cti=cti)
        serial_trap_list, serial_ccd = self._serial_traps_ccd_from(cti=cti)

        self.check_traps(trap_list_0=parallel_trap_list, trap_list_1=serial_trap_list)
        self.check_ccd(ccd_list=[parallel_ccd, serial_ccd])

        parallel_ccd = self.ccd_from(ccd_phase=parallel_ccd)
        serial_ccd = self.ccd_from(ccd_phase=serial_ccd)

        try:
            image_cti_removed = remove_cti(
                image=data,
                n_iterations=self.iterations,
                parallel_ccd=parallel_ccd,
                parallel_roe=self.parallel_roe,
                parallel_traps=parallel_trap_list,
                parallel_express=self.parallel_express,
                parallel_window_offset=data.readout_offsets[0],
                parallel_window_start=self.parallel_window_start,
                parallel_window_stop=self.parallel_window_stop,
                parallel_time_start=self.parallel_time_start,
                parallel_time_stop=self.parallel_time_stop,
                parallel_prune_n_electrons=self.parallel_prune_n_electrons,
                parallel_prune_frequency=self.parallel_prune_frequency,
                serial_ccd=serial_ccd,
                serial_roe=self.serial_roe,
                serial_traps=serial_trap_list,
                serial_express=self.serial_express,
                serial_window_offset=data.readout_offsets[1],
                serial_window_start=self.serial_window_start,
                serial_window_stop=self.serial_window_stop,
                serial_time_start=self.serial_time_start,
                serial_time_stop=self.serial_time_stop,
                serial_prune_n_electrons=self.serial_prune_n_electrons,
                serial_prune_frequency=self.serial_prune_frequency,
                allow_negative_pixels=self.allow_negative_pixels,
                pixel_bounce=self.pixel_bounce,
            )
        except TypeError:
            image_cti_removed = remove_cti(
                image=data,
                n_iterations=self.iterations,
                parallel_ccd=parallel_ccd,
                parallel_roe=self.parallel_roe,
                parallel_traps=parallel_trap_list,
                parallel_express=self.parallel_express,
                parallel_window_offset=data.readout_offsets[0],
                parallel_window_start=self.parallel_window_start,
                parallel_window_stop=self.parallel_window_stop,
                parallel_time_start=self.parallel_time_start,
                parallel_time_stop=self.parallel_time_stop,
                parallel_prune_n_electrons=self.parallel_prune_n_electrons,
                parallel_prune_frequency=self.parallel_prune_frequency,
                serial_ccd=serial_ccd,
                serial_roe=self.serial_roe,
                serial_traps=serial_trap_list,
                serial_express=self.serial_express,
                serial_window_offset=data.readout_offsets[1],
                serial_window_start=self.serial_window_start,
                serial_window_stop=self.serial_window_stop,
                serial_time_start=self.serial_time_start,
                serial_time_stop=self.serial_time_stop,
                serial_prune_n_electrons=self.serial_prune_n_electrons,
                serial_prune_frequency=self.serial_prune_frequency,
                pixel_bounce=self.pixel_bounce,
            )

        return aa.Array2D(
            values=image_cti_removed,
            mask=data.mask,
            header=data.header,
            store_native=True,
            skip_mask=True,
        )



from typing import List

from autoarray.structures.arrays.uniform_2d import Array2D


def master_ci_from(ci_list: List[Array2D]) -> Array2D:
    """
    Determine the master charge injection frame from a list of observations of charge injection images, which all use
    the same charge injection parameters and therefore should all be identical except for read noise.

    The master frame is computed by summing all images in the list and dividing by the number of images, e.g. it is
    the mean value across the images in every pixel.

    There are many effects this function does not account for (bias, NL, cosmics, CTI, etc.) that will most likely
    need to be accounted for before a more realistic implementation is possible.

    Parameters
    ----------
    ci_list
        A list of charge injection images all of which are taken using the same charge injection
        parameters / electronics.

    Returns
    -------
    The estimated master charge injection image.
    """

    master_ci = sum(ci_list) / len(ci_list)

    return Array2D.no_mask(
        values=master_ci.native, pixel_scales=ci_list[0].pixel_scales
    )



from typing import List

import autoarray as aa

from autocti.extract.two_d.parallel.abstract import Extract2DParallel
from autocti.extract.settings import SettingsExtract

from autocti.extract.two_d import extract_2d_util


class Extract2DParallelOverscan(Extract2DParallel):
    def region_list_from(self, settings: SettingsExtract) -> List[aa.Region2D]:
        """
        Returns a list of the 2D parallel overscan region, which is the parallel overscan input to the
        object, between two input `pixels` indexes (this is somewhat redundant information, but mimicks
        the `Extract` object API across all other `Extract` objects).

        (top-row, bottom-row, left-column, right-column) = (y0, y1, x0, x1)

        The parallel overscan spans all columns of the image, thus the coordinates x0 and x1 do not change. y0 and y1
        are updated based on the `pixels` input.

        Negative pixel values can be input into the `pixels` tuple, whereby rows in front of the parallel overscan are
        also extracted.

        The diagram below illustrates the extraction for `pixels=(0, 1)`:

        [] = read-out electronics
        [==========] = read-out register
        [..........] = serial prescan
        [pppppppppp] = parallel overscan
        [ssssssssss] = serial overscan
        [f#ff#f#f#f] = signal region (FPR) (0 / 1 indicate the region index)
        [tttttttttt] = parallel / serial charge injection region trail

               [ppppppppppppppppppppp]
               [ppppppppppppppppppppp]
           [...][ttttttttttttttttttttt][sss]
           [...][c1c1cc1c1cc1cc1ccc1cc][sss]
        |  [...][1c1c1cc1c1cc1ccc1cc1][sss]    |
        |  [...][ttttttttttttttttttttt][sss]    | Direction
        Par [...][ttttttttttttttttttttt][sss]    | of
        |  [...][0ccc0cccc0cccc0cccc0c][sss]    | clocking
        |/  [...][cc0ccc0cccc0cccc0cccc][sss]    \/

        []     [=====================]
               <---------Ser--------

        The extracted regions correspond to the parallel overscan [ppppppppppppppp] regions.

        For `pixels=(0,1)` the extracted arrays returned via the `array_2d_list_from()` function keep the first
        parallel pixels across the rows of the overscan:

        array_2d_list[0] = [ppppppppppppppppppppp]

        Parameters
        ----------
        settings
           The settings used to extract the parallel overscan, which for example include the `pixels` tuple specifying the
           range of pixel rows they are extracted between.
        """

        pixels = settings.pixels

        if settings.pixels_from_end is not None:
            pixels = (
                self.parallel_overscan.total_rows - settings.pixels_from_end,
                self.parallel_overscan.total_rows,
            )

        parallel_overscan_extract = aa.Region2D(
            (
                self.parallel_overscan.y0 + pixels[0],
                self.parallel_overscan.y0 + pixels[1],
                self.parallel_overscan.x0,
                self.parallel_overscan.x1,
            )
        )

        return [parallel_overscan_extract]

    def binned_region_1d_from(self, settings: SettingsExtract) -> aa.Region1D:
        """
        The `Extract` objects allow one to extract a `Dataset1D` from a 2D CTI dataset, which is used to perform
        CTI modeling in 1D.

        This is performed by binning up the data via the `binned_array_1d_from` function.

        In order to create the 1D dataset a `Layout1D` is required, which requires the `region_list` containing the
        charge regions on the 1D dataset (e.g. where the FPR appears in 1D after binning).

        The function returns the this region if the 1D dataset is extracted from the parallel overscan. This
        assumes the overscan contains EPER trails and therefore all pixels in front of the overscan act as the
        charge region and therefore FPR. This is the case when science imaging flat field data is used.

        The charge region is therefore only included if there are negative entries in the `pixels` tuple, meaning that
        pixels before the overscan (e.g. the FPR) are extracted.

        Parameters
        ----------
        settings
           The settings used to extract the parallel overscan, which for example include the `pixels` tuple specifying the
           range of pixel rows they are extracted between.
        """
        return extract_2d_util.binned_region_1d_eper_from(pixels=settings.pixels)



import numpy as np
from typing import List, Optional, Tuple

import autoarray as aa

from autocti.extract.two_d.abstract import Extract2D
from autocti.extract.settings import SettingsExtract


class Extract2DSerial(Extract2D):
    @property
    def binning_axis(self) -> int:
        """
        The axis over which binning is performed to turn 2D serial data (E.g. an EPER) into 1D data.

        For a serial extract `axis=0` such that binning is performed over the rows containing the EPER.
        """
        return 0

    def _value_list_from(
        self, array: aa.Array2D, value_str: str, settings: SettingsExtract
    ):
        value_list = []

        arr_list = [
            array.native[region.slice]
            for region in self.region_list_from(settings=settings)
        ]

        mask_list = [
            array.mask[region.slice]
            for region in self.region_list_from(settings=settings)
        ]

        arr_stack = np.stack(arr_list)
        mask_stack = np.stack(mask_list)

        for row_index in range(arr_list[0].shape[0]):
            arr_extract = arr_stack[:, row_index, :]
            mask_extract = mask_stack[:, row_index, :]

            arr_unmasked = arr_extract[np.invert(mask_extract)]

            if value_str == "median":
                value_list.append(float(np.median(arr_unmasked)))
            elif value_str == "mean":
                value_list.append(float(np.mean(arr_unmasked)))
            elif value_str == "std":
                value_list.append(float(np.std(arr_unmasked)))

        return value_list

    def median_list_from(
        self, array: aa.Array2D, settings: SettingsExtract
    ) -> List[float]:
        """
        Returns the median values of the `Extract2D` object's corresponding region, where the median is taken over
        all rows of the region(s).

        To describe this function we will use the example of estimating the median of the charge lines in charge
        injection data, by taking the median over the rows of charge injection (e.g. each median is anti-aligned with
        the FPR).

        By taking the median of values of the charge injection regions (after accounting for those which have had
        electrons captured and relocated due to CTI), we can therefore estimate whether there is row-to-row variation
        in the charge injection.

        This function does this for every rows of every charge injection. If multiple charge injections are performed,
        the median of all charge regions is taken.

        For example, if there are 3 charge injection regions, this function returns a list where each value
        estimates the charge injection normalization of a given charge injection region. The size of this list
        is therefore the number of charge injection rows.

        The function `median_list_of_lists_from` performs the median over each individual charge injection
        region in each row. It therefore estimates multiple injection normalizations per row. for each individual
        charge region. Which function one uses depends on the properties of the charge injection on the instrumentation.

        Parameters
        ----------
        settings
           The settings used to extract the serial region (e.g. the EPERs), which for example include the `pixels`
           tuple specifying the range of pixel columns they are extracted between.
        """
        return self._value_list_from(
            array=array,
            value_str="median",
            settings=settings,
        )

    def _value_list_of_lists_from(
        self, array: aa.Array2D, value_str: str, settings: SettingsExtract
    ):
        value_lists = []

        arr_list = [
            array.native[region.slice]
            for region in self.region_list_from(settings=settings)
        ]

        mask_list = [
            array.mask[region.slice]
            for region in self.region_list_from(settings=settings)
        ]

        for array_2d, mask in zip(arr_list, mask_list):
            value_list = []

            for row_index in range(array_2d.shape[0]):
                arr_extract = array_2d[row_index, :]
                mask_extract = mask[row_index, :]

                arr_unmasked = arr_extract[np.invert(mask_extract)]

                if value_str == "median":
                    value = float(np.median(arr_unmasked))
                elif value_str == "mean":
                    value = float(np.mean(arr_unmasked))
                elif value_str == "std":
                    value = float(np.std(arr_unmasked))

                value_list.append(value)

            value_lists.append(value_list)

        return value_lists

    def median_list_of_lists_from(
        self, array: aa.Array2D, settings: SettingsExtract
    ) -> List[List]:
        """
        Returns the median values of the `Extract2D` object's corresponding region, where the median is taken over
        all row(s) of every individual region.

        To describe this function we will use the example of estimating the median of the charge lines in charge
        injection data, by taking the median over every individual row of charge injection (e.g. aligned with the
        FPR).

        The inner regions of the FPR of each charge injection line informs us of the injected level of
        charge for that injection.

        By taking the median of values of the charge injection regions (after accounting for those which have had
        electrons captured and relocated due to CTI), we can therefore estimate the input charge injection
        normalization.

        This function does this for every row of every individual charge injection for every charge injection region.

        For example, if there are 3 charge injection regions, this function returns a list of lists where the outer
        list contains 3 lists each of which give estimates of the charge injection normalization in a given charge
        injection region. Thd size of the inner list is therefore the number of charge injection rows.

        The function `median_list_from` performs the median over all charge injection region in each
        row and thus estimates a single injection normalization per row. Which function one uses depends on the
        properties of the charge injection on the instrumentation.

        Parameters
        ----------
        settings
           The settings used to extract the serial region (e.g. the EPERs), which for example include the `pixels`
           tuple specifying the range of pixel columns they are extracted between.
        """
        return self._value_list_of_lists_from(
            array=array,
            value_str="median",
            settings=settings,
        )



from autoconf.dictable import from_dict, from_json, output_to_json, to_dict

from autoarray.mask.mask_1d import Mask1D
from autoarray.layout.region import Region1D
from autoarray.layout.region import Region2D
from autoarray.structures.arrays.uniform_1d import Array1D
from autoarray.structures.arrays.uniform_2d import Array2D
from autoarray.structures.header import Header
from autoarray.dataset import preprocess
from autoarray.dataset.imaging.imaging import Imaging

from arcticpy import ROE
from arcticpy import ROEChargeInjection
from arcticpy import CCD
from arcticpy import CCDPhase
from arcticpy import TrapInstantCapture
from arcticpy import TrapSlowCapture
from arcticpy import TrapInstantCaptureContinuum
from arcticpy import TrapSlowCaptureContinuum

from .charge_injection.fit import FitImagingCI
from .charge_injection.hyper import HyperCINoiseScalar
from .charge_injection.hyper import HyperCINoiseCollection
from .charge_injection.imaging.imaging import ImagingCI
from .charge_injection.imaging.settings import SettingsImagingCI
from .charge_injection.imaging.simulator import SimulatorImagingCI
from .charge_injection.layout import Layout2DCI
from .cosmics.cosmics import SimulatorCosmicRayMap
from .extract.settings import SettingsExtract
from .extract.two_d.parallel.overscan import Extract2DParallelOverscan
from .extract.two_d.parallel.fpr import Extract2DParallelFPR
from .extract.two_d.parallel.eper import Extract2DParallelEPER
from .extract.two_d.parallel.pre_injection import Extract2DParallelPreInjection
from .extract.two_d.parallel.pedestal import Extract2DParallelPedestal
from .extract.two_d.serial.prescan import Extract2DSerialPrescan
from .extract.two_d.serial.overscan import Extract2DSerialOverscan
from .extract.two_d.serial.overscan_no_eper import Extract2DSerialOverscanNoEPER
from .extract.two_d.serial.fpr import Extract2DSerialFPR
from .extract.two_d.serial.eper import Extract2DSerialEPER
from .extract.two_d.parallel.calibration import Extract2DParallelCalibration
from .extract.two_d.serial.calibration import Extract2DSerialCalibration
from .extract.two_d.master import Extract2DMaster
from .instruments import euclid
from .instruments import acs
from .charge_injection.fit import FitImagingCI
from .mask.mask_2d import Mask2D
from .mask.mask_2d import SettingsMask2D
from .mask.mask_1d import Mask1D
from .mask.mask_1d import SettingsMask1D
from .extract.one_d.overscan import Extract1DOverscan
from .extract.one_d.fpr import Extract1DFPR
from .extract.one_d.eper import Extract1DEPER
from .extract.one_d.master import Extract1DMaster
from .layout.one_d import Layout1D
from .dataset_1d.dataset_1d.settings import SettingsDataset1D
from .dataset_1d.dataset_1d.dataset_1d import Dataset1D
from .dataset_1d.dataset_1d.simulator import SimulatorDataset1D
from .dataset_1d.fit import FitDataset1D
from .dataset_1d.model.analysis import AnalysisDataset1D
from .dataset_1d.model.result import ResultDataset1D
from .charge_injection.model.analysis import AnalysisImagingCI
from .charge_injection.model.result import ResultImagingCI
from .model.model_util import CTI1D
from .model.model_util import CTI2D
from .model.settings import SettingsCTI1D
from .model.settings import SettingsCTI2D
from .clocker.one_d import Clocker1D
from .clocker.two_d import Clocker2D
from . import aggregator as agg
from . import util
from . import plot
from . import mock as m  # noqa

from autoconf import conf

conf.instance.register(__file__)

__version__ = "2023.10.23.2"



from copy import deepcopy
from typing import Tuple

import autoarray as aa

from autocti.charge_injection.layout import Layout2DCI
from autocti.charge_injection.imaging.imaging import ImagingCI
from autocti.mask.mask_2d import Mask2D


class Extract2DParallelCalibration:
    def __init__(self, shape_2d: Tuple[int, int], region_list: aa.type.Region2DList):
        """
        Class containing methods for extracting a parallel calibration dataset from a 2D CTI calibration dataset.

        The parallel calibration region is the region of a dataset that is necessary for fitting a parallel-only CTI
        model. For example, for charge injection imaging, parallel EPERs form only in columns of the CCD where charge
        is injected, the serial prescan and overscan have no signal. The parallel calibration dataset therefore
        extracts only these columns.

        A subset of the parallel calibration data may also be extracted (e.g. only the first row of every charge region)
        for fast initial CTI modeling.

        This uses the `region_list`, which contains the regions with charge (e.g. the charge injection regions) in
        pixel coordinates.

        Parameters
        ----------
        shape_2d
            The two dimensional shape of the charge injection imaging, corresponding to the number of rows (pixels
            in parallel direction) and columns (pixels in serial direction).
        region_list
            Integer pixel coordinates specifying the corners of each charge region (top-row, bottom-row,
            left-column, right-column).
        """
        self.shape_2d = shape_2d
        self.region_list = list(map(aa.Region2D, region_list))

    def extraction_region_from(self, columns: Tuple[int, int]) -> aa.type.Region2DLike:
        """
         Extract a parallel calibration array from an input array, where this array contains a sub-set of the input
         array which is specifically used for only parallel CTI calibration. This array is simply a specified number
         of columns that are closest to the read-out electronics.

         The diagram below illustrates the arrays that is extracted from a array with columns=(0, 3):

         [] = read-out electronics
         [==========] = read-out register
         [..........] = serial prescan
         [pppppppppp] = parallel overscan
         [ssssssssss] = serial overscan
         [f#ff#f#f#f] = signal region (FPR) (0 / 1 indicate the region index)
         [tttttttttt] = parallel / serial charge injection region trail

                [ptptptptptptptptptptp]
                [tptptptptptptptptptpt]
           [...][xxxxxxxxxxxxxxxxxxxxx][sss]
           [...][ccccccccccccccccccccc][sss]
         | [...][ccccccccccccccccccccc][sts]    |
         | [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | Direction
         P [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | of
         | [...][ccccccccccccccccccccc][sss]    | clocking
         \/[...][ccccccccccccccccccccc][sss]   \/

         []     [=====================]
                <--------Ser---------

         The extracted array keeps just the trails following all charge injection scans and replaces all other
         values with 0s:

                [ptp]
                [tpt]
                [xxx]
                [ccc]
         |      [ccc]                           |
         |      [xxx]                           | Direction
        Par     [xxx]                           | of
         |      [ccc]                           | clocking
                [ccc]                           |

         []     [=====================]
                <--------Ser---------
        """
        return self.region_list[0].serial_towards_roe_full_region_from(
            shape_2d=self.shape_2d, pixels=columns
        )

    def mask_2d_from(self, mask: aa.Mask2D, columns: Tuple[int, int]) -> Mask2D:
        """
        Extract a mask to go with a parallel calibration array from an input mask.

        The parallel calibration array is described in the function `array_2d_from()`.
        """
        extraction_region = self.extraction_region_from(columns=columns)

        return Mask2D(
            mask=mask[extraction_region.slice], pixel_scales=mask.pixel_scales
        )

    def array_2d_from(self, array: aa.Array2D, columns: Tuple[int, int]) -> aa.Array2D:
        """
        Extract a parallel calibration array from an input array, where this array contains a sub-set of the input
        array which is specifically used for only parallel CTI calibration. This array is simply a specified number
        of columns that are closest to the read-out electronics.

        The diagram below illustrates the arrays that is extracted from a array with columns=(0, 3):

        [] = read-out electronics
        [==========] = read-out register
        [..........] = serial prescan
        [pppppppppp] = parallel overscan
        [ssssssssss] = serial overscan
        [f#ff#f#f#f] = signal region (FPR) (0 / 1 indicate the region index)
        [tttttttttt] = parallel / serial charge injection region trail

               [ptptptptptptptptptptp]
               [tptptptptptptptptptpt]
          [...][xxxxxxxxxxxxxxxxxxxxx][sss]
          [...][ccccccccccccccccccccc][sss]
        | [...][ccccccccccccccccccccc][sts]    |
        | [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | Direction
        P [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | of
        | [...][ccccccccccccccccccccc][sss]    | clocking
        \/[...][ccccccccccccccccccccc][sss]   \/

        []     [=====================]
               <--------Ser---------

        The extracted array keeps just the trails following all charge injection scans and replaces all other
        values with 0s:

               [ptp]
               [tpt]
               [xxx]
               [ccc]
        |      [ccc]                           |
        |      [xxx]                           | Direction
        P      [xxx]                           | of
        |      [ccc]                           | clocking
        |/     [ccc]                          \/

        []     [=====================]
               <--------Ser---------
        """
        extraction_region = self.extraction_region_from(columns=columns)
        mask_2d = self.mask_2d_from(mask=array.mask, columns=columns)

        return aa.Array2D(
            values=array.native[extraction_region.slice],
            mask=mask_2d,
            header=array.header,
        )

    def extracted_layout_from(self, layout, columns: Tuple[int, int]) -> Layout2DCI:
        """
        Extract the layout of a parallel calibration array from an input layout, where this layout contains the regions
        of the input layout which are retained after the parallel CTI calibration array is created. This layout
        is simply a specified number of columns that are closest to the read-out electronics.

        The diagram below illustrates the arrays that is extracted from a array with columns=(0, 3):

        [] = read-out electronics
        [==========] = read-out register
        [..........] = serial prescan
        [pppppppppp] = parallel overscan
        [ssssssssss] = serial overscan
        [f#ff#f#f#f] = signal region (FPR) (0 / 1 indicate the region index)
        [tttttttttt] = parallel / serial charge injection region trail

               [ptptptptptptptptptptp]
               [tptptptptptptptptptpt]
          [...][xxxxxxxxxxxxxxxxxxxxx][sss]
          [...][ccccccccccccccccccccc][sss]
        | [...][ccccccccccccccccccccc][sts]    |
        | [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | Direction
        P [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | of
        | [...][ccccccccccccccccccccc][sss]    | clocking
        \/[...][ccccccccccccccccccccc][sss]    \/

        []     [=====================]
               <--------Ser---------

        The extracted array keeps just the trails following all charge injection scans and replaces all other
        values with 0s:

               [ptp]
               [tpt]
               [xxx]
               [ccc]
        |      [ccc]                           |
        |      [xxx]                           | Direction
        P      [xxx]                           | of
        |      [ccc]                           | clocking
        |/     [ccc]                           \/

        []     [=====================]
               <--------Ser---------
        """

        extraction_region = self.extraction_region_from(columns=columns)

        return self.with_extracted_regions(
            layout=layout, extraction_region=extraction_region
        )

    def with_extracted_regions(
        self, layout, extraction_region: aa.type.Region2DLike
    ) -> Layout2DCI:
        layout = deepcopy(layout)

        extracted_region_list = list(
            map(
                lambda region: aa.util.layout.region_after_extraction(
                    original_region=region, extraction_region=extraction_region
                ),
                self.region_list,
            )
        )
        extracted_region_list = list(filter(None, extracted_region_list))
        if not extracted_region_list:
            extracted_region_list = None

        layout.region_list = extracted_region_list
        return layout

    def imaging_ci_from(
        self, dataset: ImagingCI, columns: Tuple[int, int]
    ) -> ImagingCI:
        """
        Returnss a function to extract a parallel section for given columns
        """

        from autocti.charge_injection.imaging.imaging import ImagingCI

        cosmic_ray_map = (
            dataset.layout.extract.parallel_calibration.array_2d_from(
                array=dataset.cosmic_ray_map, columns=columns
            )
            if dataset.cosmic_ray_map is not None
            else None
        )

        if dataset.noise_scaling_map_dict is not None:
            noise_scaling_map_dict = {
                key: dataset.layout.extract.parallel_calibration.array_2d_from(
                    array=noise_scaling_map, columns=columns
                )
                for key, noise_scaling_map in dataset.noise_scaling_map_dict.items()
            }

        else:
            noise_scaling_map_dict = None

        extraction_region = (
            dataset.layout.extract.parallel_calibration.extraction_region_from(
                columns=columns
            )
        )

        mask = self.mask_2d_from(mask=dataset.mask, columns=columns)

        dataset = ImagingCI(
            data=dataset.layout.extract.parallel_calibration.array_2d_from(
                array=dataset.data, columns=columns
            ),
            noise_map=dataset.layout.extract.parallel_calibration.array_2d_from(
                array=dataset.noise_map, columns=columns
            ),
            pre_cti_data=dataset.layout.extract.parallel_calibration.array_2d_from(
                array=dataset.pre_cti_data, columns=columns
            ),
            layout=dataset.layout.layout_extracted_from(
                extraction_region=extraction_region
            ),
            cosmic_ray_map=cosmic_ray_map,
            noise_scaling_map_dict=noise_scaling_map_dict,
            fpr_value=dataset.fpr_value,
            settings_dict=dataset.settings_dict,
        )

        return dataset.apply_mask(mask=mask)



from os import path
from autoconf import conf
from autoarray.plot.wrap.base import Output
from autoarray.plot.mat_plot.one_d import MatPlot1D
from autoarray.plot.mat_plot.two_d import MatPlot2D
from autoarray.plot.include.one_d import Include1D
from autoarray.plot.include.two_d import Include2D


def setting(section, name):
    return conf.instance["visualize"]["plots"][section][name]


def plot_setting(section, name):
    return setting(section, name)


class Visualizer:
    def __init__(self, visualize_path):
        self.visualize_path = visualize_path

        self.include_1d = Include1D()
        self.include_2d = Include2D()

    def mat_plot_1d_from(self, subfolders, format="png"):
        return MatPlot1D(
            output=Output(
                path=path.join(self.visualize_path, subfolders), format=format
            )
        )

    def mat_plot_2d_from(self, subfolders, format="png"):
        return MatPlot2D(
            output=Output(
                path=path.join(self.visualize_path, subfolders), format=format
            )
        )



from typing import Optional, Tuple

import autoarray as aa

from autocti.extract.settings import SettingsExtract
from autocti.extract.two_d.parallel.overscan import Extract2DParallelOverscan
from autocti.extract.two_d.parallel.fpr import Extract2DParallelFPR
from autocti.extract.two_d.parallel.eper import Extract2DParallelEPER
from autocti.extract.two_d.parallel.pedestal import Extract2DParallelPedestal
from autocti.extract.two_d.parallel.pre_injection import Extract2DParallelPreInjection
from autocti.extract.two_d.serial.overscan import Extract2DSerialOverscan
from autocti.extract.two_d.serial.prescan import Extract2DSerialPrescan
from autocti.extract.two_d.serial.fpr import Extract2DSerialFPR
from autocti.extract.two_d.serial.eper import Extract2DSerialEPER
from autocti.extract.two_d.serial.overscan_no_eper import Extract2DSerialOverscanNoEPER
from autocti.extract.two_d.parallel.calibration import Extract2DParallelCalibration
from autocti.extract.two_d.serial.calibration import Extract2DSerialCalibration


class Extract2DMaster:
    def __init__(
        self,
        region_list,
        shape_2d: Optional[Tuple[int, int]] = None,
        parallel_overscan: Optional[aa.type.Region2DLike] = None,
        serial_prescan: Optional[aa.type.Region2DLike] = None,
        serial_overscan: Optional[aa.type.Region2DLike] = None,
    ):
        """
        Class which groups all `Extract` classes, which are classes containing methods for extracting specific
        regions from 2D CTI calibration data (e.g. the FPRs of a charge injection image)

        This uses the `region_list`, which contains the regions with input known charge on the CTI calibration
        data in pixel coordinates.

        Parameters
        ----------
        shape_2d
            The two dimensional shape of the charge injection imaging, corresponding to the number of rows (pixels
            in parallel direction) and columns (pixels in serial direction).
        region_list
            Integer pixel coordinates specifying the corners of each charge injection region (top-row, bottom-row,
            left-column, right-column).
        parallel_overscan
            Integer pixel coordinates specifying the corners of the parallel overscan (top-row, bottom-row,
            left-column, right-column).
        serial_prescan
            Integer pixel coordinates specifying the corners of the serial prescan (top-row, bottom-row,
            left-column, right-column).
        serial_overscan
            Integer pixel coordinates specifying the corners of the serial overscan (top-row, bottom-row,
            left-column, right-column).
        """

        self.region_list = (
            list(map(aa.Region2D, region_list)) if region_list is not None else None
        )

        self.shape_2d = shape_2d
        self._parallel_overscan = parallel_overscan
        self._serial_prescan = serial_prescan
        self._serial_overscan = serial_overscan

    @property
    def parallel_fpr(self):
        return Extract2DParallelFPR(
            shape_2d=self.shape_2d,
            region_list=self.region_list,
            parallel_overscan=self._parallel_overscan,
            serial_prescan=self._serial_prescan,
            serial_overscan=self._serial_overscan,
        )

    @property
    def parallel_eper(self):
        return Extract2DParallelEPER(
            shape_2d=self.shape_2d,
            region_list=self.region_list,
            parallel_overscan=self._parallel_overscan,
            serial_prescan=self._serial_prescan,
            serial_overscan=self._serial_overscan,
        )

    @property
    def parallel_pedestal(self):
        return Extract2DParallelPedestal(
            shape_2d=self.shape_2d,
            region_list=self.region_list,
            parallel_overscan=self._parallel_overscan,
            serial_prescan=self._serial_prescan,
            serial_overscan=self._serial_overscan,
        )

    @property
    def parallel_pre_injection(self):
        return Extract2DParallelPreInjection(
            shape_2d=self.shape_2d,
            region_list=self.region_list,
            parallel_overscan=self._parallel_overscan,
            serial_prescan=self._serial_prescan,
            serial_overscan=self._serial_overscan,
        )

    @property
    def parallel_overscan(self):
        return Extract2DParallelOverscan(parallel_overscan=self._parallel_overscan)

    @property
    def parallel_calibration(self):
        return Extract2DParallelCalibration(
            shape_2d=self.shape_2d, region_list=self.region_list
        )

    @property
    def serial_fpr(self):
        return Extract2DSerialFPR(
            shape_2d=self.shape_2d,
            region_list=self.region_list,
            parallel_overscan=self._parallel_overscan,
            serial_prescan=self._serial_prescan,
            serial_overscan=self._serial_overscan,
        )

    @property
    def serial_eper(self):
        return Extract2DSerialEPER(
            shape_2d=self.shape_2d,
            region_list=self.region_list,
            parallel_overscan=self._parallel_overscan,
            serial_prescan=self._serial_prescan,
            serial_overscan=self._serial_overscan,
        )

    @property
    def serial_calibration(self):
        return Extract2DSerialCalibration(
            shape_2d=self.shape_2d,
            region_list=self.region_list,
            serial_prescan=self._serial_prescan,
            serial_overscan=self._serial_overscan,
        )

    @property
    def serial_prescan(self):
        return Extract2DSerialPrescan(serial_prescan=self._serial_prescan)

    @property
    def serial_overscan(self):
        return Extract2DSerialOverscan(serial_overscan=self._serial_overscan)

    @property
    def serial_overscan_no_eper(self):
        return Extract2DSerialOverscanNoEPER(
            shape_2d=self.shape_2d,
            region_list=self.region_list,
            parallel_overscan=self._parallel_overscan,
            serial_prescan=self._serial_prescan,
            serial_overscan=self._serial_overscan,
        )

    def regions_array_2d_from(self, array: aa.Array2D) -> aa.Array2D:
        """
          Extract all of the charge-injection regions from an input `Array2D` object and returns them as a new `Array2D`
          where these extracted regions are included and all other entries are zeros.

          The dimensions of the input array therefore do not change (unlike other ``Extract`` methods).

          The diagram below illustrates the extraction:

          [] = read-out electronics
          [==========] = read-out register
          [..........] = serial prescan
          [pppppppppp] = parallel overscan
          [ssssssssss] = serial overscan
          [f#ff#f#f#f] = signal region (FPR) (0 / 1 indicate the region index)
          [tttttttttt] = parallel / serial charge injection region trail

                 [ppppppppppppppppppppp]
                 [ppppppppppppppppppppp]
            [...][xxxxxxxxxxxxxxxxxxxxx][sss]
            [...][ccccccccccccccccccccc][sss]
          | [...][ccccccccccccccccccccc][sss]    |
          | [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | Direction
        Par [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | of
          | [...][ccccccccccccccccccccc][sss]    | clocking
          /  [...][ccccccccccccccccccccc][sss]   \/

          []     [=====================]
                 <--------Ser---------

          The extracted array keeps just the charge injection region, all other values become 0:

                 [000000000000000000000]
                 [000000000000000000000]
            [000][000000000000000000000][000]
            [000][ccccccccccccccccccccc][000]
          | [000][ccccccccccccccccccccc][000]    |
          | [000][000000000000000000000][000]    | Direction
         Par[000][000000000000000000000][000]    | of
          | [000][ccccccccccccccccccccc][000]    | clocking
         |/ [000][ccccccccccccccccccccc][000]   \/

          []     [=====================]
                 <--------Ser---------
        """

        new_array = array.native.copy() * 0.0

        for region in self.region_list:
            new_array[region.slice] += array.native[region.slice]

        return new_array

    def non_regions_array_2d_from(self, array: aa.Array2D) -> aa.Array2D:
        """
         Extract all of the areas of an `Array2D` that are not within any of the layout's charge-injection regions
         and return them as a new `Array2D` where these extracted regions are included and the charge injection regions
         are zeros

         The extracted array therefore includes all EPER trails and other regions of the image which may contain
         signal but are not in the FPR.

         The dimensions of the input array therefore do not change (unlike other ``Extract`` methods).

         The diagram below illustrates the extraction:

         [] = read-out electronics
         [==========] = read-out register
         [..........] = serial prescan
         [pppppppppp] = parallel overscan
         [ssssssssss] = serial overscan
         [f#ff#f#f#f] = signal region (FPR) (0 / 1 indicate the region index)
         [tttttttttt] = parallel / serial charge injection region trail

                [tptpptptptpptpptpptpt]
                [tptptptpptpttptptptpt]
           [...][ttttttttttttttttttttt][sss]
           [...][ccccccccccccccccccccc][sss]
         | [...][ccccccccccccccccccccc][sss]    |
         | [...][ttttttttttttttttttttt][sss]    | Direction
        Par[...][ttttttttttttttttttttt][sss]    | of
         | [...][ccccccccccccccccccccc][sss]    | clocking
         |/ [...][ccccccccccccccccccccc][sss]   \/

         []     [=====================]
                <--------Ser---------

         The extracted array keeps everything except the charge injection  region,which become 0s:

                [tptpptptptpptpptpptpt]
                [tptptptpptpttptptptpt]
           [000][ttttttttttttttttttttt][000]
           [000][000000000000000000000][000]
         | [000][000000000000000000000][000]    |
         | [000][ttttttttttttttttttttt][000]    | Direction
        Par[000][ttttttttttttttttttttt][000]    | of
         | [000][000000000000000000000][000]    | clocking
           [000][000000000000000000000][000]   \/

         []     [=====================]
                <--------Ser---------
        """

        non_regions_ci_array = array.native.copy()

        for region in self.region_list:
            non_regions_ci_array[region.slice] = 0.0

        return non_regions_ci_array

    def parallel_fpr_and_eper_array_2d_from(
        self,
        array: aa.Array2D,
        fpr_pixels: Tuple[int, int] = None,
        eper_pixels: Tuple[int, int] = None,
    ) -> aa.Array2D:
        """
            Extract all of the data values in an input `array2D` corresponding to the parallel front edges and trails of
            each the charge-injection region.

            One can specify the range of rows that are extracted, for example:

            fpr_pixels = (0, 1) will extract just the first leading front edge row.
            fpr_pixels = (0, 2) will extract the leading two front edge rows.
            trails_pixels = (0, 1) will extract the first row of trails closest to the charge injection region.

            The diagram below illustrates the arrays that are extracted from the input array for `fpr_pixels=(0,1)`
            and `trails_pixels=(0,1)`:

            The diagram below illustrates the extraction:

            [] = read-out electronics
            [==========] = read-out register
            [..........] = serial prescan
            [pppppppppp] = parallel overscan
            [ssssssssss] = serial overscan
            [f#ff#f#f#f] = signal region (FPR) (0 / 1 indicate the region index)
            [tttttttttt] = parallel / serial charge injection region trail

                   [tptpptptptpptpptpptpt]
                   [tptptptpptpttptptptpt]
              [...][ttttttttttttttttttttt][sss]
              [...][ccccccccccccccccccccc][sss]
            | [...][ccccccccccccccccccccc][sss]    |
            | [...][ttttttttttttttttttttt][sss]    | Direction
            P [...][ttttttttttttttttttttt][sss]    | of
            | [...][ccccccccccccccccccccc][sss]    | clocking
              [...][ccccccccccccccccccccc][sss]    |

            []     [=====================]
                   <--------Ser---------

            The extracted array keeps just the leading edges and trails following all charge injection scans and
            replaces all other values with 0s:

                   [000000000000000000000]
                   [000000000000000000000]
              [000][ttttttttttttttttttttt][000]
              [000][000000000000000000000][000]
            | [000][ccccccccccccccccccccc][000]    |
            | [000][000000000000000000000][000]    | Direction
            P [000][ttttttttttttttttttttt][000]    | of
            | [000][000000000000000000000][000]    | clocking
              [000][ccccccccccccccccccccc][000]    |

            []     [=====================]
                   <--------Ser---------

            Parameters
        ----------
            fpr_pixels
                The row indexes to extract the front edge between (e.g. rows(0, 3) extracts the 1st, 2nd and 3rd rows).
            trails_pixels
                The row indexes to extract the trails between (e.g. rows(0, 3) extracts the 1st, 2nd and 3rd rows)
        """
        new_array = array.native.copy() * 0.0

        if fpr_pixels is not None:
            new_array = self.parallel_fpr.add_to_array(
                new_array=new_array,
                array=array,
                settings=SettingsExtract(pixels=fpr_pixels),
            )

        if eper_pixels is not None:
            new_array = self.parallel_eper.add_to_array(
                new_array=new_array,
                array=array,
                settings=SettingsExtract(pixels=eper_pixels),
            )

        return new_array

    def serial_fpr_and_eper_array_2d_from(
        self, array: aa.Array2D, fpr_pixels=None, eper_pixels=None
    ) -> aa.Array2D:
        """
            Extract an array of all of the serial FPRs and EPERs of each the charge-injection scans from a charge
            injection array.

            One can specify the range of columns that are extracted, for example:

            fpr_pixels = (0, 1) will extract just the leading front edge column.
            fpr_pixels = (0, 2) will extract the leading two front edge columns.
            trails_pixels = (0, 1) will extract the first column of trails closest to the charge injection region.

            The diagram below illustrates the arrays that is extracted from a array for `fpr_pixels=(0,2)` and
            `trails_pixels=(0,2)`:

            [] = read-out electronics
            [==========] = read-out register
            [..........] = serial prescan
            [pppppppppp] = parallel overscan
            [ssssssssss] = serial overscan
            [f#ff#f#f#f] = signal region (FPR) (0 / 1 indicate the region index)
            [tttttttttt] = parallel / serial charge injection region trail

                   [ppppppppppppppppppppp]
                   [ppppppppppppppppppppp]
              [...][xxxxxxxxxxxxxxxxxxxxx][sss]
              [...][ccccccccccccccccccccc][sts]
            | [...][ccccccccccccccccccccc][tst]    |
            | [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | Direction
            P [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | of
            | [...][ccccccccccccccccccccc][sts]    | clocking
              [...][ccccccccccccccccccccc][tst]    |

            []     [=====================]
                   <--------Ser---------

            The extracted array keeps just the leading edge and trails following all charge injection scans and
            replaces all other values with 0s:

                   [000000000000000000000]
                   [000000000000000000000]
              [000][000000000000000000000][000]
              [000][cc0000000000000000000][st0]
            | [000][cc0000000000000000000][ts0]    |
            | [000][000000000000000000000][000]    | Direction
            P [000][000000000000000000000][000]    | of
            | [000][cc0000000000000000000][st0]    | clocking
              [000][cc0000000000000000000][st0]    |

            []     [=====================]
                   <--------Ser---------

            Parameters
        ----------
            array
            fpr_pixels
                The column indexes to extract the front edge between (e.g. columns(0, 3) extracts the 1st, 2nd and 3rd rows)
            trails_pixels
                The column indexes to extract the trails between (e.g. columns(0, 3) extracts the 1st, 2nd and 3rd rows)
        """
        new_array = array.native.copy() * 0.0

        if fpr_pixels is not None:
            new_array = self.serial_fpr.add_to_array(
                new_array=new_array,
                array=array,
                settings=SettingsExtract(pixels=fpr_pixels),
            )

        if eper_pixels is not None:
            new_array = self.serial_eper.add_to_array(
                new_array=new_array,
                array=array,
                settings=SettingsExtract(pixels=eper_pixels),
            )

        return new_array

    def serial_overscan_above_eper_array_2d_from(self, array: aa.Array2D) -> aa.Array2D:
        """
        Extract an array of the region above the EPER trails of the serial overscan.

        This region does not contain signal from either parallel or serial EPERs, however it may have a faint signal
        from charge that is trailed from the parallel EPER's in the serial direction during serial clocking.

        [] = read-out electronics
        [==========] = read-out register
        [..........] = serial prescan
        [pppppppppp] = parallel overscan
        [ssssssssss] = serial overscan
        [f#ff#f#f#f] = signal region (FPR) (0 / 1 indicate the region index)
        [tttttttttt] = parallel / serial charge injection region trail

               [ppppppppppppppppppppp]
               [ppppppppppppppppppppp]
          [...][xxxxxxxxxxxxxxxxxxxxx][sss]
          [...][ccccccccccccccccccccc][tst]
        | [...][ccccccccccccccccccccc][sts]    |
        | [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | Direction
        P [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | of
        | [...][ccccccccccccccccccccc][tst]    | clocking
          [...][ccccccccccccccccccccc][sts]    |

        []     [=====================]
               <--------Ser---------

        The extracted array keeps just the trails following all charge injection scans and replaces all other
        values with 0s:

               [000000000000000000000]
               [000000000000000000000]
          [000][000000000000000000000][sss]
          [000][000000000000000000000][000]
        | [000][000000000000000000000][000]    |
        | [000][000000000000000000000][sss]    | Direction
        P [000][000000000000000000000][sss]    | of
        | [000][000000000000000000000][000]    | clocking
          [000][000000000000000000000][000]    |

        []     [=====================]
               <--------Ser---------
        """
        serial_overscan = self.serial_overscan.serial_overscan

        new_array = array.native.copy() * 0.0

        new_array[serial_overscan.slice] = array.native[serial_overscan.slice]

        trails_region_list = list(
            map(
                lambda ci_region: ci_region.serial_trailing_region_from(
                    (0, serial_overscan.total_columns)
                ),
                self.region_list,
            )
        )

        for region in trails_region_list:
            new_array[region.slice] = 0

        return new_array



from autoarray.layout.layout import Layout2D
from autoarray.layout.region import Region2D


class Layout2DEuclid(Layout2D):
    @classmethod
    def from_fits_header(cls, ext_header):
        """
        Use an input array of a Euclid quadrant and its corresponding .fits file header to rotate the quadrant to
        the correct orientation for arCTIc clocking.

        See the docstring of the `Array2DEuclid` class for a complete description of the Euclid FPA, quadrants and
        rotations.
        """

        ccd_id = ext_header["CCDID"]
        quadrant_id = ext_header["QUADID"]

        parallel_overscan_size = ext_header.get("OVRSCANY", default=None)
        if parallel_overscan_size is None:
            parallel_overscan_size = 0
        serial_overscan_size = ext_header.get("OVRSCANX", default=None)
        serial_prescan_size = ext_header.get("PRESCANX", default=None)
        serial_size = ext_header.get("NAXIS1", default=None)
        parallel_size = ext_header.get("NAXIS2", default=None)

        return cls.from_ccd_and_quadrant_id(
            ccd_id=ccd_id,
            quadrant_id=quadrant_id,
            parallel_size=parallel_size,
            serial_size=serial_size,
            parallel_overscan_size=parallel_overscan_size,
            serial_prescan_size=serial_prescan_size,
            serial_overscan_size=serial_overscan_size,
        )

    @classmethod
    def from_ccd_and_quadrant_id(
        cls,
        ccd_id: str,
        quadrant_id: str,
        parallel_size: int = 2086,
        serial_size: int = 2128,
        serial_prescan_size: int = 51,
        serial_overscan_size: int = 29,
        parallel_overscan_size: int = 20,
    ):
        """
        Use an input array of a Euclid quadrant, its ccd_id and quadrant_id  to rotate the quadrant to
        the correct orientation for arCTIc clocking.

        See the docstring of the `Array2DEuclid` class for a complete description of the Euclid FPA, quadrants and
        rotations.
        """

        row_index = ccd_id[-1]

        if (row_index in "123") and (quadrant_id == "E"):
            return Layout2DEuclid.bottom_left(
                parallel_size=parallel_size,
                serial_size=serial_size,
                serial_prescan_size=serial_prescan_size,
                serial_overscan_size=serial_overscan_size,
                parallel_overscan_size=parallel_overscan_size,
            )
        elif (row_index in "123") and (quadrant_id == "F"):
            return Layout2DEuclid.bottom_right(
                parallel_size=parallel_size,
                serial_size=serial_size,
                serial_prescan_size=serial_prescan_size,
                serial_overscan_size=serial_overscan_size,
                parallel_overscan_size=parallel_overscan_size,
            )
        elif (row_index in "123") and (quadrant_id == "G"):
            return Layout2DEuclid.top_right(
                parallel_size=parallel_size,
                serial_size=serial_size,
                serial_prescan_size=serial_prescan_size,
                serial_overscan_size=serial_overscan_size,
                parallel_overscan_size=parallel_overscan_size,
            )
        elif (row_index in "123") and (quadrant_id == "H"):
            return Layout2DEuclid.top_left(
                parallel_size=parallel_size,
                serial_size=serial_size,
                serial_prescan_size=serial_prescan_size,
                serial_overscan_size=serial_overscan_size,
                parallel_overscan_size=parallel_overscan_size,
            )
        elif (row_index in "456") and (quadrant_id == "E"):
            return Layout2DEuclid.top_right(
                parallel_size=parallel_size,
                serial_size=serial_size,
                serial_prescan_size=serial_prescan_size,
                serial_overscan_size=serial_overscan_size,
                parallel_overscan_size=parallel_overscan_size,
            )
        elif (row_index in "456") and (quadrant_id == "F"):
            return Layout2DEuclid.top_left(
                parallel_size=parallel_size,
                serial_size=serial_size,
                serial_prescan_size=serial_prescan_size,
                serial_overscan_size=serial_overscan_size,
                parallel_overscan_size=parallel_overscan_size,
            )
        elif (row_index in "456") and (quadrant_id == "G"):
            return Layout2DEuclid.bottom_left(
                parallel_size=parallel_size,
                serial_size=serial_size,
                serial_prescan_size=serial_prescan_size,
                serial_overscan_size=serial_overscan_size,
                parallel_overscan_size=parallel_overscan_size,
            )
        elif (row_index in "456") and (quadrant_id == "H"):
            return Layout2DEuclid.bottom_right(
                parallel_size=parallel_size,
                serial_size=serial_size,
                serial_prescan_size=serial_prescan_size,
                serial_overscan_size=serial_overscan_size,
                parallel_overscan_size=parallel_overscan_size,
            )

    @classmethod
    def top_left(
        cls,
        parallel_size: int = 2086,
        serial_size: int = 2128,
        serial_prescan_size: int = 51,
        serial_overscan_size: int = 29,
        parallel_overscan_size: int = 20,
    ):
        """
        Load the layout of a Euclid quadrant where the read out electronics (ROE) are at the top left of the CCD.

        When the corresponding array of data is loaded, it is rotated to ensure that parallel and serial CTI clocking
        go in the correct direction for arctic clocking.

        This function first species the layout regions on the unrotated image array (e.g. the parallel overscan indexes
        are defined at the bottom of the data). The layout is updated to account for this rotation (e.g. for
        a top left CCD the parallel overscan indexes are flipped to the top of the array).

        ROE's in the `top_left` correspond to 1/4 of the 144 quadrants on a Euclid FPA, specifically:

        - Those in the top half of the FPA (CCDID = 1, 2 or 3) and quadrant id H.
        - Those in the bottom half of the FPA (CCDID = 4, 5 or 6) and quadrant id F.

        The data is rotated by a flip-up-down, such that the ROE electrons are in the bottom left corner after rotation.

        See the docstring of the `Array2DEuclid` class for a complete description of the Euclid FPA, quadrants and
        rotations.

        Parameters
        ----------
        array_electrons
            The array of data (in units of electrons) which has the ROE in the top left and is rotated for arctic
            clocking.
        ccd_id
            The ID of the euclid CCD (1, 2, 3, 4, 5, or 6) indicating which half of the FP the CCD is in.
        quadrant_id
            The ID of the quadrant (E, G, F or H).
        """
        if parallel_overscan_size > 0:
            parallel_overscan = Region2D(
                (
                    0,
                    parallel_overscan_size,
                    serial_prescan_size,
                    serial_size - serial_overscan_size,
                )
            )

        else:
            parallel_overscan = None

        serial_prescan = Region2D((0, parallel_size, 0, serial_prescan_size))
        serial_overscan = Region2D(
            (
                0,
                parallel_size - parallel_overscan_size,
                serial_size - serial_overscan_size,
                serial_size,
            )
        )

        layout_2d = Layout2DEuclid(
            shape_2d=(parallel_size, serial_size),
            original_roe_corner=(0, 0),
            parallel_overscan=parallel_overscan,
            serial_prescan=serial_prescan,
            serial_overscan=serial_overscan,
        )

        return layout_2d.new_rotated_from(roe_corner=(0, 0))

    @classmethod
    def top_right(
        cls,
        parallel_size: int = 2086,
        serial_size: int = 2128,
        serial_prescan_size: int = 51,
        serial_overscan_size: int = 29,
        parallel_overscan_size: int = 20,
    ):
        """
        Load the layout of a Euclid quadrant where the read out electronics (ROE) are at the top right of the CCD.

        When the corresponding array of data is loaded, it is rotated to ensure that parallel and serial CTI clocking
        go in the correct direction for arctic clocking.

        This function first species the layout regions on the unrotated image array (e.g. the serial prescan indexes
        are defined on the right-hand size of the data). The layout is updated to account for this rotation (e.g. the
        serial prescan indexes now become defined on the left-hand side).

        ROE's in the `top_right` correspond to 1/4 of the 144 quadrants on a Euclid FPA, specifically:

        - Those in the top half of the FPA (CCDID = 1, 2 or 3) and quadrant id G.
        - Those in the bottom half of the FPA (CCDID = 4, 5 or 6) and quadrant id E.

        The data is rotated by a flip-up-down and flip left-right, such that the ROE electrons are in the bottom left
        corner after rotation.

        See the docstring of the `Array2DEuclid` class for a complete description of the Euclid FPA, quadrants and
        rotations.

        Parameters
        ----------
        parallel_size
            The size (number of pixels) of the array across the parallel direction.
        serial_size
            The size (number of pixels) of the array across the serial direction.
        serial_prescan_size
            The size (number of pixels) of the serial prescan.
        serial_overscan_size
            The size (number of pixels) of the serial overscan.
        parallel_overscan_size
            The size (number of pixels) of the parallel overscan.
        """
        if parallel_overscan_size > 0:
            parallel_overscan = Region2D(
                (
                    0,
                    parallel_overscan_size,
                    serial_overscan_size,
                    serial_size - serial_prescan_size,
                )
            )

        else:
            parallel_overscan = None

        serial_prescan = Region2D(
            (0, parallel_size, serial_size - serial_prescan_size, serial_size)
        )
        serial_overscan = Region2D(
            (0, parallel_size - parallel_overscan_size, 0, serial_overscan_size)
        )

        layout_2d = Layout2DEuclid(
            shape_2d=(parallel_size, serial_size),
            original_roe_corner=(0, 1),
            parallel_overscan=parallel_overscan,
            serial_prescan=serial_prescan,
            serial_overscan=serial_overscan,
        )

        return layout_2d.new_rotated_from(roe_corner=(0, 1))

    @classmethod
    def bottom_left(
        cls,
        parallel_size: int = 2086,
        serial_size: int = 2128,
        serial_prescan_size: int = 51,
        serial_overscan_size: int = 29,
        parallel_overscan_size: int = 20,
    ):
        """
        Load the layout of a Euclid quadrant where the read out electronics (ROE) are at the bottom left of the CCD.

        When the corresponding array of data is loaded, it is rotated to ensure that parallel and serial CTI clocking
        go in the correct direction for arctic clocking.

        This function first species the layout regions on the unrotated image array (e.g. the serial prescan indexes
        are defined on the left-hand size of the data). The layout is updated to account for this rotation (e.g. for
        a bottom left CCD no rotation is performed, thus the serial prescan indexes stay on the left-hand side).

        ROE's in the `bottom left` correspond to 1/4 of the 144 quadrants on a Euclid FPA, specifically:

        - Those in the top half of the FPA (CCDID = 1, 2 or 3) and quadrant id E.
        - Those in the bottom half of the FPA (CCDID = 4, 5 or 6) and quadrant id G.

        No rotation is performed for arCTIc clocking because the code assume the ROE are at the bottom left.

        See the docstring of the `Array2DEuclid` class for a complete description of the Euclid FPA, quadrants and
        rotations.

        Parameters
        ----------
        parallel_size
            The size (number of pixels) of the array across the parallel direction.
        serial_size
            The size (number of pixels) of the array across the serial direction.
        serial_prescan_size
            The size (number of pixels) of the serial prescan.
        serial_overscan_size
            The size (number of pixels) of the serial overscan.
        parallel_overscan_size
            The size (number of pixels) of the parallel overscan.
        """
        if parallel_overscan_size > 0:
            parallel_overscan = Region2D(
                (
                    parallel_size - parallel_overscan_size,
                    parallel_size,
                    serial_prescan_size,
                    serial_size - serial_overscan_size,
                )
            )

        else:
            parallel_overscan = None

        serial_prescan = Region2D((0, parallel_size, 0, serial_prescan_size))
        serial_overscan = Region2D(
            (
                0,
                parallel_size - parallel_overscan_size,
                serial_size - serial_overscan_size,
                serial_size,
            )
        )

        layout_2d = Layout2DEuclid(
            shape_2d=(parallel_size, serial_size),
            original_roe_corner=(1, 0),
            parallel_overscan=parallel_overscan,
            serial_prescan=serial_prescan,
            serial_overscan=serial_overscan,
        )

        return layout_2d.new_rotated_from(roe_corner=(1, 0))

    @classmethod
    def bottom_right(
        cls,
        parallel_size: int = 2086,
        serial_size: int = 2128,
        serial_prescan_size: int = 51,
        serial_overscan_size: int = 29,
        parallel_overscan_size: int = 20,
    ):
        """
        Load the layout of a Euclid quadrant where the read out electronics (ROE) are at the bottom right of the CCD.

        When the corresponding array of data is loaded, it is rotated to ensure that parallel and serial CTI clocking
        go in the correct direction for arctic clocking.

        This function first species the layout regions on the unrotated image array (e.g. the serial prescan indexes
        are defined on the right-hand size of the data). The layout is updated to account for this rotation (e.g. the
        serial prescan indexes now become defined on the left-hand side).

        ROE's in the `bottom right` correspond to 1/4 of the 144 quadrants on a Euclid FPA, specifically:

        - Those in the top half of the FPA (CCDID = 1, 2 or 3) and quadrant id F.
        - Those in the bottom half of the FPA (CCDID = 4, 5 or 6) and quadrant id H.

        The data and layout coordinates are rotated by a flip left-right, such that the ROE electrons are in the
        bottom left corner after rotation.

        See the docstring of the `Array2DEuclid` class for a complete description of the Euclid FPA, quadrants and
        rotations.

        Parameters
        ----------
        parallel_size
            The size (number of pixels) of the array across the parallel direction.
        serial_size
            The size (number of pixels) of the array across the serial direction.
        serial_prescan_size
            The size (number of pixels) of the serial prescan.
        serial_overscan_size
            The size (number of pixels) of the serial overscan.
        parallel_overscan_size
            The size (number of pixels) of the parallel overscan.
        """
        if parallel_overscan_size > 0:
            parallel_overscan = Region2D(
                (
                    parallel_size - parallel_overscan_size,
                    parallel_size,
                    serial_overscan_size,
                    serial_size - serial_prescan_size,
                )
            )

        else:
            parallel_overscan = None

        serial_prescan = Region2D(
            (0, parallel_size, serial_size - serial_prescan_size, serial_size)
        )
        serial_overscan = Region2D(
            (0, parallel_size - parallel_overscan_size, 0, serial_overscan_size)
        )

        layout_2d = Layout2DEuclid(
            shape_2d=(parallel_size, serial_size),
            original_roe_corner=(1, 1),
            parallel_overscan=parallel_overscan,
            serial_prescan=serial_prescan,
            serial_overscan=serial_overscan,
        )

        return layout_2d.new_rotated_from(roe_corner=(1, 1))



import numpy as np
import math
from typing import Dict, List, Optional, Tuple
import autoarray as aa

from autocti.layout.two_d import Layout2D


def generate_column(size: int, norm: float, row_slope: float) -> np.ndarray:
    """
    Generate a column of non-uniform charge, including row non-uniformity.

    The pixel-numbering used to generate non-uniformity across the charge injection rows runs from 1 -> size

    Parameters
    ----------
    size
        The size of the non-uniform column of charge
    normalization
        The input normalization of the column's charge e.g. the level of charge injected.

    """
    return norm * (np.arange(1, size + 1)) ** row_slope


def region_ci_from(
    region_dimensions: Tuple[int, int],
    injection_norm_list: List[float],
    row_slope: Optional[float] = 0.0,
) -> np.ndarray:
    """
    Generate a non-uniform charge injection region from an input list of normalization values across columns.
    """

    ci_rows = region_dimensions[0]
    ci_region = np.zeros(region_dimensions)

    for column_index, injection_norm in enumerate(injection_norm_list):
        ci_region[0:ci_rows, column_index] = generate_column(
            size=ci_rows, norm=injection_norm, row_slope=row_slope
        )

    return ci_region


def region_list_ci_via_electronics_from(
    injection_start: int,
    injection_on: int,
    injection_off: int,
    injection_total: int,
    parallel_size: int,
    serial_size: int,
    serial_prescan_size: int,
    serial_overscan_size: int,
    roe_corner: Tuple[int, int],
):
    region_list_ci = []

    for index in range(injection_total):
        if roe_corner == (0, 0):
            ci_region = (
                parallel_size - (injection_start + injection_on),
                parallel_size - injection_start,
                serial_prescan_size,
                serial_size - serial_overscan_size,
            )

        elif roe_corner == (1, 0):
            ci_region = (
                injection_start,
                injection_start + injection_on,
                serial_prescan_size,
                serial_size - serial_overscan_size,
            )

        elif roe_corner == (0, 1):
            ci_region = (
                parallel_size - (injection_start + injection_on),
                parallel_size - injection_start,
                serial_overscan_size,
                serial_size - serial_prescan_size,
            )

        elif roe_corner == (1, 1):
            ci_region = (
                injection_start,
                injection_start + injection_on,
                serial_overscan_size,
                serial_size - serial_prescan_size,
            )

        region_list_ci.append(ci_region)

        injection_start += injection_on + injection_off

    return region_list_ci



import astropy.io.fits as pyfits
import numpy as np
from os import path
from scipy.interpolate import interp1d
from typing import Tuple

import autoarray as aa


class SimulatorCosmicRayMap:
    def __init__(
        self,
        shape_native: Tuple[int, int],
        lengths: np.ndarray,
        distances: np.ndarray,
        flux_scaling: float = 1.0,
        pixel_scale: float = 0.1,
        seed=-1,
    ):
        """
        Returns a map of cosmic rays.

        Adapted from the OU-SIM ELVIS Simulator in Euclid, originally written by Patrick Hudelot.

        Parameters
        ----------
        shape_native
            The 2D shape of the cosmic ray map, corresponding to the image the cosmic arrays are to be added to.
        flux_scaling
            A factor which scales the overall normalization of the cosmic rays.
        settings_dict
            A dictionary of all settings that control the behaviour of the cosmic ray simulator.
        seed
            Random number seed, set to positive value for reproduceable cosmic ray maps.
        """

        self.shape_native = shape_native

        self.lengths = lengths
        self.distances = distances

        self.settings_dict = {}
        self.flux_scaling = flux_scaling

        self.pixel_scale = pixel_scale

        if seed == -1:
            seed = np.random.randint(
                0, 1e9
            )  # Use one seed, so all regions have identical column non-uniformity.
        np.random.seed(seed)

    @classmethod
    def from_fits(
        cls,
        length_file: str,
        distance_file: str,
        shape_native: Tuple[int, int],
        flux_scaling: float = 1.0,
        pixel_scale: float = 0.1,
        seed=-1,
    ) -> "SimulatorCosmicRayMap":
        """
        Creates a cosmic ray simulator where the length and distance arrays are loaded from user supplied .fits files.

        Parameters
        ----------
        length_file
            A .fits table describing the lengths properties of the cosmic rays that are simulated.
        distance_file
            A .fits table describing the distance properties of the cosmic rays that are simulated.
        shape_native
            The 2D shape of the cosmic ray map, corresponding to the image the cosmic arrays are to be added to.
        flux_scaling
            A factor which scales the overall normalization of the cosmic rays.
        """

        lengths = pyfits.getdata(length_file, extname="DATA")
        lengths = np.array(lengths.tolist())

        distances = pyfits.getdata(distance_file, extname="DATA")
        distances = np.array(distances.tolist())

        return SimulatorCosmicRayMap(
            shape_native=shape_native,
            lengths=lengths,
            distances=distances,
            flux_scaling=flux_scaling,
            pixel_scale=pixel_scale,
            seed=seed,
        )

    @classmethod
    def defaults(
        cls,
        shape_native: Tuple[int, int],
        flux_scaling: float = 1.0,
        pixel_scale: float = 0.1,
        seed=-1,
    ) -> "SimulatorCosmicRayMap":
        """
        Creates a cosmic ray simulator where the length and distance arrays are loaded from the default files stored
        on the `PyAutoCTI` repository in `autocti.cosmics`.

        Parameters
        ----------
        length_file
            A .fits table describing the lengths properties of the cosmic rays that are simulated.
        distance_file
            A .fits table describing the distance properties of the cosmic rays that are simulated.
        shape_native
            The 2D shape of the cosmic ray map, corresponding to the image the cosmic arrays are to be added to.
        flux_scaling
            A factor which scales the overall normalization of the cosmic rays.
        """

        dir_path = path.dirname(path.realpath(__file__))

        from autocti.cosmics.cr_lengths import lengths
        from autocti.cosmics.cr_distances import distances

        return SimulatorCosmicRayMap(
            shape_native=shape_native,
            lengths=lengths,
            distances=distances,
            flux_scaling=flux_scaling,
            pixel_scale=pixel_scale,
            seed=seed,
        )

    def intercepts_from(self, luminosities, x0, y0, lengths, angles, flux_scaling):
        """
        Derive cosmic ray streak intercept points.

        Parameters
        ----------
        luminosities
            The luminosities of the cosmic ray tracks.
        x0
            Central positions of the cosmic ray tracks in x-direction.
        y0
            Central positions of the cosmic ray tracks in y-direction.
        lengths
            The lengths of the cosmic ray tracks.
        angles
            The orientation angles of the cosmic ray tracks.
        """
        # create empty arrays
        image = np.zeros((self.shape_native[0], self.shape_native[1]), dtype=np.float64)

        # x and y shifts
        dx = lengths * np.cos(angles) / 2.0  # beware! 0<phi< pi, dx < 0
        dy = lengths * np.sin(angles) / 2.0
        mskdx = np.abs(dx) < 1e-8
        mskdy = np.abs(dy) < 1e-8
        dx[mskdx] = 0.0
        dy[mskdy] = 0.0

        # pixels in x-direction
        ilo = np.round(x0.copy() - dx)
        mask = ilo < 0.0
        ilo[mask] = 0
        ilo = ilo.astype(int)

        ihi = np.round(x0.copy() + dx)
        mask = ihi > self.shape_native[1]
        ihi[mask] = self.shape_native[1]
        ihi = ihi.astype(int)

        # pixels in y-directions
        jlo = np.round(y0.copy() - dy)
        mask = jlo < 0.0
        jlo[mask] = 0
        jlo = jlo.astype(int)

        jhi = np.round(y0.copy() + dy)
        mask = jhi > self.shape_native[0]
        jhi[mask] = self.shape_native[0]
        jhi = jhi.astype(int)

        offending_delta = 1.0

        # loop over the individual events
        for i, luminosity in enumerate(luminosities):
            n = 0  # count the intercepts

            u = []
            x = []
            y = []

            # Compute X intercepts on the pixel grid
            if ilo[i] < ihi[i]:
                for xcoord in range(ilo[i], ihi[i]):
                    ok = (xcoord - x0[i]) / dx[i]
                    if np.abs(ok) <= offending_delta:
                        n += 1
                        u.append(ok)
                        x.append(xcoord)
                        y.append(y0[i] + ok * dy[i])
            else:
                for xcoord in range(ihi[i], ilo[i]):
                    ok = (xcoord - x0[i]) / dx[i]
                    if np.abs(ok) <= offending_delta:
                        n += 1
                        u.append(ok)
                        x.append(xcoord)
                        y.append(y0[i] + ok * dy[i])

            # Compute Y intercepts on the pixel grid
            if jlo[i] < jhi[i]:
                for ycoord in range(jlo[i], jhi[i]):
                    ok = (ycoord - y0[i]) / dy[i]
                    if np.abs(ok) <= offending_delta:
                        n += 1
                        u.append(ok)
                        x.append(x0[i] + ok * dx[i])
                        y.append(ycoord)
            else:
                for ycoord in range(jhi[i], jlo[i]):
                    ok = (ycoord - y0[i]) / dy[i]
                    if np.abs(ok) <= offending_delta:
                        n += 1
                        u.append(ok)
                        x.append(x0[i] + ok * dx[i])
                        y.append(ycoord)

            # check if no intercepts were found
            if n < 1:
                xc = int(np.floor(x0[i]))
                yc = int(np.floor(y0[i]))
                image[yc, xc] += luminosity * flux_scaling

            # Find the arguments that sort the intersections along the track
            u = np.asarray(u)
            x = np.asarray(x)
            y = np.asarray(y)

            args = np.argsort(u)

            u = u[args]
            x = x[args]
            y = y[args]

            # Decide which cell each interval traverses, and the path length
            for j in range(1, n - 1):
                w = (u[j + 1] - u[j]) / 2.0
                cx = int(1 + np.floor((x[j + 1] + x[j]) / 2.0))
                cy = int(1 + np.floor((y[j + 1] + y[j]) / 2.0))

                if 0 <= cx < self.shape_native[1] and 0 <= cy < self.shape_native[0]:
                    image[cy, cx] += w * luminosity * flux_scaling

        return image

    def cosmic_ray_map_from(
        self, cover_fraction: float = 1.4, limit: float = 1000.0
    ) -> aa.Array2D:
        """
        Return a cosmic ray, where cosmic rays are generated using the lengths and distance of the class instance.


        Parameters
        ----------
        limit
            The limiting energy for the cosmic ray event.
        cover_fraction
            The covering fraction of cosmic rays over the total number of pixels (in percent) normalized for a 5
            65s exposure time.
        """

        # Prepare the CR map
        cosmic_ray_map = np.zeros((self.shape_native[0], self.shape_native[1]))

        # how many events to draw at once, too large number leads to exceeding the covering fraction

        cdf = self.lengths[:, 1]
        ucr = self.lengths[:, 0]
        approx_pdf = (cdf[1:] - cdf[0:-1]) / (ucr[1:] - ucr[0:-1])
        average_length = (approx_pdf * ucr[1:]).sum() / approx_pdf.sum()
        total_guess = (
            cover_fraction
            / 100.0
            * self.shape_native[1]
            * self.shape_native[0]
            / average_length
        )

        # allocating for a max. 5% error in cover. fraction, aprox.
        # Notice that the minimum number of events will be one...
        cr_n = max(int(total_guess * 0.05), 1)

        covering = 0.0
        lengths = []
        energies = []
        total_cosmics = 0

        while covering < cover_fraction:
            # pseudo-random numbers taken from a uniform distribution between 0 and 1
            luck = np.random.rand(cr_n)

            # draw the length of the tracks
            try:
                ius = interp1d(self.lengths[:, 1], self.lengths[:, 0], kind="slinear")
            except ValueError:
                ius = interp1d(self.lengths[:, 1], self.lengths[:, 0], kind="linear")
            length = ius(luck)

            if limit is None:
                ius = interp1d(
                    self.distances[:, 1], self.distances[:, 0], kind="slinear"
                )
                energy = ius(luck)
            else:
                # set the energy directly to the limit
                energy = np.asarray([limit])

            lengths += length.tolist()
            energies += energy.tolist()

            # Choose the properties such as positions and an angle from a random Uniform dist
            x = self.shape_native[1] * np.random.rand(int(np.floor(cr_n)))
            y = self.shape_native[0] * np.random.rand(int(np.floor(cr_n)))
            angle = np.pi * np.random.rand(int(np.floor(cr_n)))

            # find the intercepts
            cosmic_ray_map += self.intercepts_from(
                energy, x, y, length, angle, self.flux_scaling
            )

            # count the covering factor
            area = np.count_nonzero(cosmic_ray_map)
            covering = 100.0 * area / (self.shape_native[1] * self.shape_native[0])

            total_cosmics += cr_n

        return aa.Array2D.no_mask(values=cosmic_ray_map, pixel_scales=self.pixel_scale)



import numpy as np
from typing import Optional, List, Tuple, Union

import autoarray as aa

from autocti.charge_injection.imaging.imaging import ImagingCI
from autocti.dataset_1d.dataset_1d.dataset_1d import Dataset1D
from autocti.extract.settings import SettingsExtract
from autocti.layout.one_d import Layout1D


class Extract2D:
    def __init__(
        self,
        shape_2d: Optional[Tuple[int, int]] = None,
        region_list: Optional[aa.type.Region2DList] = None,
        parallel_overscan: Optional[aa.type.Region2DLike] = None,
        serial_prescan: Optional[aa.type.Region2DLike] = None,
        serial_overscan: Optional[aa.type.Region2DLike] = None,
    ):
        """
        Abstract class containing methods for extracting regions from a 2D charge injection image.

        This uses the `region_list`, which contains the charge injection regions in pixel coordinates.

        Parameters
        ----------
        region_list
            Integer pixel coordinates specifying the corners of each charge injection region (top-row, bottom-row,
            left-column, right-column).
        """

        self.shape_2d = shape_2d

        self.region_list = (
            list(map(aa.Region2D, region_list)) if region_list is not None else None
        )

        self.parallel_overscan = (
            aa.Region2D(region=parallel_overscan)
            if isinstance(parallel_overscan, tuple)
            else parallel_overscan
        )
        self.serial_prescan = (
            aa.Region2D(region=serial_prescan)
            if isinstance(serial_prescan, tuple)
            else serial_prescan
        )
        self.serial_overscan = (
            aa.Region2D(region=serial_overscan)
            if isinstance(serial_overscan, tuple)
            else serial_overscan
        )

    @property
    def pedestal(self) -> aa.Region2D:
        """
        The region of the charge injection image containing the pedestal.
        """
        return aa.Region2D(
            region=(
                self.parallel_overscan.y0,
                self.shape_2d[0],
                self.serial_overscan.x0,
                self.shape_2d[1],
            )
        )

    @property
    def total_rows_min(self) -> int:
        """
        The number of rows between the read-out electronics and the charge injection region closest to them.
        """
        return np.min([region.total_rows for region in self.region_list])

    @property
    def total_columns_min(self) -> int:
        """
        The number of columns between the read-out electronics and the charge injection region closest to them.
        """
        return np.min([region.total_columns for region in self.region_list])

    @property
    def binning_axis(self) -> int:
        """
        The axis over which binning is performed to turn a 2D FPR into a 1D FPR.

        For a parallel extract `axis=1` such that binning is performed over the rows containing the FPR.
        """
        raise NotImplementedError

    @property
    def parallel_rows_between_regions(self) -> List[int]:
        """
        Returns a list where each entry is the number of pixels a charge injection region and its neighboring
        charge injection region.
        """
        return [
            self.region_list[i + 1].y0 - self.region_list[i].y1
            for i in range(len(self.region_list) - 1)
        ]

    @property
    def parallel_rows_to_array_edge(self) -> int:
        """
        The number of pixels from the edge of the parallel EPERs to the edge of the array.

        This is the number of pixels from the last charge injection FPR edge to the read out register and electronics
        and will include the parallel overscan if the CCD has one.
        """
        return self.shape_2d[0] - np.max([region.y1 for region in self.region_list])

    def region_list_from(self, settings: SettingsExtract) -> List[aa.Region2D]:
        raise NotImplementedError

    @property
    def anti_region_list(self) -> List[aa.Region2D]:
        """
        The `region_list` contains the charge injection FPR regions in pixel coordinates.

        The `anti_region_list` contains the regions between the charge injection FPRs in pixel coordinates.

        These predominantly contain the parallel EPERs, but may also contain the region in front of the FPR which do
        not contain any EPERs because no charge has been injected yet.

        The `anti_region_list` also does not include the serial prescan and serial overscan regions, but does include
        the parallel overscan region.

        Returns
        -------
        The regions in front of and between the charge injection FPRs in pixel coordinates.
        """
        anti_region_list = []

        x0 = self.region_list[0].x0
        x1 = self.region_list[0].x1

        region_pre_fpr = aa.Region2D(region=(0, self.region_list[0].y0, x0, x1))

        anti_region_list.append(region_pre_fpr)

        if len(self.region_list) > 1:
            for i in range(len(self.region_list) - 1):
                region_between_fprs = aa.Region2D(
                    region=(self.region_list[i].y1, self.region_list[i + 1].y0, x0, x1)
                )

                anti_region_list.append(region_between_fprs)

        region_post_final_fpr = aa.Region2D(
            region=(self.region_list[-1].y1, self.shape_2d[0], x0, x1)
        )

        anti_region_list.append(region_post_final_fpr)

        return anti_region_list

    def array_2d_list_from(
        self, array: aa.Array2D, settings: SettingsExtract
    ) -> List[aa.Array2D]:
        """
        Extract a specific region from every signal region (e.g. the charge injection region of charge injection data)
        on the CTI calibration data and return as a list of 2D arrays.

        For example, this might extract the parallel EPERs of every charge injection region.

        The `region_list_from()` of each `Extract2D` class describes the exact extraction performed for each
        extract when this function is called.

        Parameters
        ----------
        array
            The array from which the regions are extracted and put into the returned list of arrays.
        settings
           The settings used to extract the serial region (e.g. the EPERs), which for example include the `pixels`
           tuple specifying the range of pixel columns they are extracted between.
        """
        region_list = self.region_list_from(settings=settings)
        region_list = settings.region_list_from(region_list=region_list)

        arr_list = [array.native[region.slice] for region in region_list]
        mask_2d_list = [array.mask[region.slice] for region in region_list]

        return [
            aa.Array2D(values=arr, mask=mask_2d).native
            for arr, mask_2d in zip(arr_list, mask_2d_list)
        ]

    def stacked_array_2d_from(
        self, array: aa.Array2D, settings: SettingsExtract
    ) -> np.ndarray:
        """
        Extract a region (e.g. the parallel FPR) of every signal region (e.g. the charge injection region of charge
        injection data) on the CTI calibration data and stack them by taking their mean.

        This returns the 2D average of the extracted regions (e.g. the parallel FPRs) of all of the charge injection
        regions, which for certain CCD charge injection electronics one may expect to be similar.

        For fits to charge injection data this function is also used to create images like the stacked 2D residuals,
        which therefore quantify the goodness-of-fit of a CTI model.

        Parameters
        ----------
        array
            The 2D array which contains the charge injection image from which the regions (e.g. the parallel FPRs)
            are extracted and stacked.
        settings
           The settings used to extract the serial region (e.g. the EPERs), which for example include the `pixels`
           tuple specifying the range of pixel columns they are extracted between.
        """

        arr_list = np.asarray(
            [
                np.array(array.native[region.slice])
                for region in self.region_list_from(settings=settings)
            ]
        )

        mask_2d_list = [
            array.mask[region.slice]
            for region in self.region_list_from(settings=settings)
        ]

        stacked_array_2d = np.mean(
            arr_list, axis=0, where=np.invert(np.asarray(mask_2d_list))
        )

        return aa.Array2D(
            values=np.asarray(stacked_array_2d.data),
            mask=sum(mask_2d_list) == len(mask_2d_list),
        ).native

    def stacked_array_2d_total_pixels_from(
        self, array: aa.Array2D, settings: SettingsExtract
    ) -> np.ndarray:
        """
        The function `stacked_array_2d_from` extracts a region (e.g. the parallel FPR) of every charge injection
        region on the charge injection image and stacks them by taking their mean.

        If the data being stacked is a noise-map, we need to know how many pixels were used in the stacking of every
        final pixel on the stacked 2d array in order to compute the new noise map via quadrature.

        Parameters
        ----------
        array
            The 2D array which contains the charge injection image from which the regions (e.g. the parallel FPRs)
            are extracted and stacked.
        settings
           The settings used to extract the serial region (e.g. the EPERs), which for example include the `pixels`
           tuple specifying the range of pixel columns they are extracted between.
        """

        mask_2d_list = [
            array.mask[region.slice]
            for region in self.region_list_from(settings=settings)
        ]

        arr_total_pixels = sum([np.invert(mask_2d) for mask_2d in mask_2d_list])

        return aa.Array2D(
            values=np.asarray(arr_total_pixels),
            mask=sum(mask_2d_list) == len(mask_2d_list),
        ).native

    def binned_array_1d_from(
        self, array: aa.Array2D, settings: SettingsExtract
    ) -> aa.Array1D:
        """
        Extract a region (e.g. the parallel FPR) of every signal region (e.g. the charge injection region of
        charge injection data) on the CTI calibration data, stack them by taking their mean and then bin them up to
        a 1D region (e.g. the 1D parallel FPR) by taking the mean across the direction opposite to
        clocking (e.g. bin over the serial direction for a parallel FPR).

        This returns the 1D average region (e.g. of the parallel FPR) of all of the charge injection regions. When
        binning a uniform charge injection this binning process removes noise to clearly reveal the FPR or EPER.
        For non-uniform injections this will provide an average FPR or EPER.

        For fits to charge injection data this function is also used to create images like the stacked 1D residuals,
        which therefore quantify the goodness-of-fit of a CTI model.

        Parameters
        ----------
        array
            The 2D array which contains the charge injeciton image from which the parallel FPRs are extracted and
            stacked.
        settings
           The settings used to extract the serial region (e.g. the EPERs), which for example include the `pixels`
           tuple specifying the range of pixel columns they are extracted between.
        """

        region_list = self.region_list_from(settings=settings)
        region_list = settings.region_list_from(region_list=region_list)

        arr_list = np.asarray([array.native[region.slice] for region in region_list])

        mask_list = np.asarray([array.mask[region.slice] for region in region_list])

        stacked_array_2d = np.mean(arr_list, axis=0, where=np.invert(mask_list))

        binned_array_1d = np.mean(
            stacked_array_2d,
            axis=self.binning_axis,
            where=np.invert(np.isnan(stacked_array_2d)),
        )
        return aa.Array1D.no_mask(
            values=binned_array_1d, pixel_scales=array.pixel_scale
        )

    def binned_array_1d_total_pixels_from(
        self, array: aa.Array2D, settings: SettingsExtract
    ) -> aa.Array1D:
        """
        The function `binned_array_1d_from` extracts a region (e.g. the parallel FPR) of every charge injection region
        on the charge injection image, stacks them by taking their mean and then bin them up to a 1D region
        (e.g. the 1D parallel FPR) by taking the mean across the direction opposite to clocking (e.g. bin over the
        serial direction for a parallel FPR).

        If the data being stacked is a noise-map, we need to know how many pixels were used in the stacking of every
        final pixel on the binned 1D array in order to compute the new noise map via quadrature.

        Parameters
        ----------
        array
            The 2D array which contains the charge injeciton image from which the parallel FPRs are extracted and
            stacked.
        settings
           The settings used to extract the serial region (e.g. the EPERs), which for example include the `pixels`
           tuple specifying the range of pixel columns they are extracted between.
        """

        mask_2d_list = [
            array.mask[region.slice]
            for region in self.region_list_from(settings=settings)
        ]

        arr_total_pixels = sum([np.invert(mask_2d) for mask_2d in mask_2d_list])

        binned_total_pixels = np.sum(
            np.asarray(arr_total_pixels), axis=self.binning_axis
        )

        return aa.Array1D.no_mask(
            values=binned_total_pixels, pixel_scales=array.pixel_scale
        )

    def binned_region_1d_from(self, settings: SettingsExtract) -> aa.Region1D:
        raise NotImplementedError

    def add_to_array(
        self, new_array: aa.Array2D, array: aa.Array2D, settings: SettingsExtract
    ) -> aa.Array2D:
        """
        Extracts the region (e.g. the parallel FPRs) from a charge injection image and adds them to a new image.

        Parameters
        ----------
        new_array
            The 2D array which the extracted parallel FPRs are added to.
        array
            The 2D array which contains the charge injection image from which the parallel FPRs are extracted.
        settings
           The settings used to extract the serial region (e.g. the EPERs), which for example include the `pixels`
           tuple specifying the range of pixel columns they are extracted between.
        """

        region_list = self.region_list_from(settings=settings)

        array_2d_list = self.array_2d_list_from(array=array, settings=settings)

        for arr, region in zip(array_2d_list, region_list):
            new_array[region.y0 : region.y1, region.x0 : region.x1] += arr

        return new_array

    def dataset_1d_from(
        self, dataset_2d: ImagingCI, settings: SettingsExtract
    ) -> Dataset1D:
        binned_data_1d = self.binned_array_1d_from(
            array=dataset_2d.data, settings=settings
        )

        binned_noise_map_1d = self.binned_array_1d_from(
            array=dataset_2d.noise_map, settings=settings
        )

        binned_noise_map_1d_total_pixels = self.binned_array_1d_total_pixels_from(
            array=dataset_2d.noise_map, settings=settings
        )

        binned_noise_map_1d /= np.sqrt(binned_noise_map_1d_total_pixels)

        binned_pre_cti_data_1d = self.binned_array_1d_from(
            array=dataset_2d.pre_cti_data,
            settings=settings,
        )

        binned_region_1d = self.binned_region_1d_from(settings=settings)

        layout_1d = Layout1D(
            shape_1d=binned_data_1d.shape_native, region_list=[binned_region_1d]
        )

        return Dataset1D(
            data=binned_data_1d,
            noise_map=binned_noise_map_1d,
            pre_cti_data=binned_pre_cti_data_1d,
            layout=layout_1d,
        )

    def add_gaussian_noise_to(
        self,
        array: aa.Array2D,
        settings: SettingsExtract,
        noise_sigma: float,
        noise_seed: int = -1,
    ) -> aa.Array2D:
        """
        Adds Gaussian noise of an input sigma value to the regions of the `Extract` object and returns the overall
        input array with this noise added.

        Parameters
        ----------
        array
            The 2D array which contains the charge injection where regions of Gaussian noise is added.
        settings
           The settings used to extract the serial region (e.g. the EPERs), which for example include the `pixels`
           tuple specifying the range of pixel columns they are extracted between.
        noise_sigma
            The sigma value (standard deviation) of the Gaussian from which noise values are drann.
        noise_seed
            The seed of the random number generator, used for the random noises maps.
        """

        region_list = self.region_list_from(settings=settings)

        array_2d_list = self.array_2d_list_from(array=array, settings=settings)

        array = array.native

        for arr, region in zip(array_2d_list, region_list):
            array[
                region.y0 : region.y1, region.x0 : region.x1
            ] = aa.preprocess.data_with_gaussian_noise_added(
                data=arr, sigma=noise_sigma, seed=noise_seed
            )

        return array



from autofit import mock
from autofit.non_linear.result import ResultsCollection


class MockResult(mock.MockResult):
    def __init__(
        self,
        samples=None,
        instance=None,
        model=None,
        analysis=None,
        search=None,
        mask=None,
        model_image=None,
        noise_scaling_map_dict_list_of_regions_ci=None,
        noise_scaling_map_dict_list_of_parallel_eper=None,
        noise_scaling_map_dict_list_of_serial_eper=None,
        noise_scaling_map_dict_list_of_serial_overscan_no_eper=None,
        use_as_hyper_dataset=False,
    ):
        super().__init__(
            samples=samples,
            instance=instance,
            model=model,
            analysis=analysis,
            search=search,
        )

        self.previous_model = model
        self.gaussian_tuples = None
        self.mask = None
        self.positions = None
        self.mask = mask
        self.model_image = model_image
        self.noise_scaling_map_dict_list_of_regions_ci = (
            noise_scaling_map_dict_list_of_regions_ci
        )
        self.noise_scaling_map_dict_list_of_parallel_eper = (
            noise_scaling_map_dict_list_of_parallel_eper
        )
        self.noise_scaling_map_dict_list_of_serial_eper = (
            noise_scaling_map_dict_list_of_serial_eper
        )
        self.noise_scaling_map_dict_list_of_serial_overscan_no_eper = (
            noise_scaling_map_dict_list_of_serial_overscan_no_eper
        )
        self.use_as_hyper_dataset = use_as_hyper_dataset

    @property
    def last(self):
        return self


class MockResults(ResultsCollection):
    def __init__(
        self,
        samples=None,
        instance=None,
        model=None,
        analysis=None,
        search=None,
        mask=None,
        model_image=None,
        noise_scaling_map_dict_list_of_regions_ci=None,
        noise_scaling_map_dict_list_of_parallel_eper=None,
        noise_scaling_map_dict_list_of_serial_eper=None,
        noise_scaling_map_dict_list_of_serial_overscan_no_eper=None,
        use_as_hyper_dataset=False,
    ):
        """
        A collection of results from previous phases. Results can be obtained using an index or the name of the phase
        from whence they came.
        """

        super().__init__()

        result = MockResult(
            samples=samples,
            instance=instance,
            model=model,
            analysis=analysis,
            search=search,
            mask=mask,
            model_image=model_image,
            noise_scaling_map_dict_list_of_regions_ci=noise_scaling_map_dict_list_of_regions_ci,
            noise_scaling_map_dict_list_of_parallel_eper=noise_scaling_map_dict_list_of_parallel_eper,
            noise_scaling_map_dict_list_of_serial_eper=noise_scaling_map_dict_list_of_serial_eper,
            noise_scaling_map_dict_list_of_serial_overscan_no_eper=noise_scaling_map_dict_list_of_serial_overscan_no_eper,
            use_as_hyper_dataset=use_as_hyper_dataset,
        )

        self.__result_list = [result]

    @property
    def last(self):
        """
        The result of the last phase
        """
        if len(self.__result_list) > 0:
            return self.__result_list[-1]
        return None

    def __getitem__(self, item):
        """
        Get the result of a previous phase by index

        Parameters
        ----------
        item: int
            The index of the result

        Returns
        -------
        result: Result
            The result of a previous phase
        """
        return self.__result_list[item]

    def __len__(self):
        return len(self.__result_list)



import numpy as np
from typing import List, Optional, Tuple

import autoarray as aa

from autocti.extract.one_d.abstract import Extract1D
from autocti.extract.settings import SettingsExtract


class Extract1DOverscan(Extract1D):
    def region_list_from(self, settings: SettingsExtract) -> List[aa.Region1D]:
        """
        Returns a list of the (x0, x1) regions containing the overscan of a 1D CTI dataset.

        These are used for extracting the overscan regions of 1D data.

        Negative pixel values can be input into the `pixels` tuple, whereby pixels in front of the FPRs  are also
        extracted.

        Parameters
        ----------
        pixels
            The row indexes to extract the overscan between (e.g. pixels(0, 3) extracts the 1st, 2nd and 3rd pixels)
        """

        pixels = settings.pixels

        if settings.pixels_from_end is not None:
            pixels = (
                self.overscan.total_pixels - settings.pixels_from_end,
                self.overscan.total_pixels,
            )

        return [
            aa.Region1D(
                region=(self.overscan.x0 + pixels[0], self.overscan.x0 + pixels[1])
            )
        ]



from copy import deepcopy
import numpy as np
from typing import Optional, Tuple

import autoarray as aa

from autocti.charge_injection.imaging.imaging import ImagingCI
from autocti.mask.mask_2d import Mask2D


class Extract2DSerialCalibration:
    def __init__(
        self,
        shape_2d: Tuple[int, int],
        region_list: aa.type.Region2DList,
        serial_prescan: Optional[aa.type.Region2DLike] = None,
        serial_overscan: Optional[aa.type.Region2DLike] = None,
    ):
        """
        Class containing methods for extracting a serial calibration dataset from a 2D CTI calibration dataset.

        The serial calibration region is the region of a dataset that is necessary for fitting a serial-only CTI
        model. For example, for charge injection imaging, serial EPERs form only in rows of the CCD where charge
        is injected and the regions in between have no signal. The serial calibration dataset therefore extracts only
        these rows.

        A subset of the serial calibration data may also be extracted (e.g. only the first row of every charge region)
        for fast initial CTI modeling.

        This uses the `region_list`, which contains the regions with charge (e.g. the charge injection regions) in
        pixel coordinates.

        Parameters
        ----------
        shape_2d
            The two dimensional shape of the charge injection imaging, corresponding to the number of rows (pixels
            in parallel direction) and columns (pixels in serial direction).
        region_list
            Integer pixel coordinates specifying the corners of each charge region (top-row, bottom-row,
            left-column, right-column).
        serial_prescan
            Integer pixel coordinates specifying the corners of the serial prescan (top-row, bottom-row,
            left-column, right-column).
        serial_overscan
            Integer pixel coordinates specifying the corners of the serial overscan (top-row, bottom-row,
            left-column, right-column).
        """
        self.shape_2d = shape_2d
        self.region_list = list(map(aa.Region2D, region_list))
        self.serial_prescan = serial_prescan
        self.serial_overscan = serial_overscan

    def array_2d_list_from(self, array: aa.Array2D):
        """
        Extract each charge injection region image for the serial calibration arrays when creating the
        """

        calibration_region_list = list(
            map(
                lambda ci_region: ci_region.parallel_full_region_from(
                    shape_2d=self.shape_2d
                ),
                self.region_list,
            )
        )
        return list(
            map(lambda region: array.native[region.slice], calibration_region_list)
        )

    def mask_2d_from(self, mask: aa.Mask2D, rows: Tuple[int, int]) -> Mask2D:
        """
        Extract a serial calibration array from a charge injection array, where this arrays is a sub-set of the
        array which can be used for serial-only calibration. Specifically, this array is all charge injection
        scans and their serial over-scan trails.

        The diagram below illustrates the arrays that is extracted from a array with column=5:

        ---KEY---
        ---------

        [] = read-out electronics   [==========] = read-out register

        [xxxxxxxxxx]                [..........] = serial prescan       [ssssssssss] = serial overscan
        [xxxxxxxxxx] = CCDPhase panel    [pppppppppp] = parallel overscan    [cccccccccc] = charge injection region
        [xxxxxxxxxx]                [tttttttttt] = parallel / serial charge injection region trail

        P = Parallel Direction      S = Serial Direction

               [ppppppppppppppppppppp]
               [pppppppppppppppppppp ]
          [...][xxxxxxxxxxxxxxxxxxxxx][sss]
          [...][ccccccccccccccccccccc][tst]
        | [...][ccccccccccccccccccccc][sts]    |
        | [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | Direction
        P [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | of
        | [...][ccccccccccccccccccccc][tst]    | clocking
          [...][ccccccccccccccccccccc][sts]    |

        []     [=====================]
               <--------Ser---------

        The extracted array keeps just the trails following all charge injection scans and replaces all other
        values with 0s:

        |                                      |
        |      [cccccccccccccccc][tst]         | Direction
        P      [cccccccccccccccc][tst]         | of
        |      [cccccccccccccccc][tst]         | clocking
               [cccccccccccccccc][tst]         |

        []     [=====================]
               <--------Ser---------
        """

        calibration_region_list = list(
            map(
                lambda ci_region: ci_region.parallel_full_region_from(
                    shape_2d=self.shape_2d
                ),
                self.region_list,
            )
        )
        calibration_masks = list(
            map(lambda region: mask[region.slice], calibration_region_list)
        )

        calibration_masks = list(
            map(lambda mask: mask[rows[0] : rows[1], :], calibration_masks)
        )
        return Mask2D(
            mask=np.concatenate(calibration_masks, axis=0),
            pixel_scales=mask.pixel_scales,
        )

    def array_2d_from(self, array: aa.Array2D, rows: Tuple[int, int]) -> aa.Array2D:
        """
        Extract a serial calibration array from a charge injection array, where this arrays is a sub-set of the
        array which can be used for serial-only calibration. Specifically, this array is all charge injection
        scans and their serial over-scan trails.

        The diagram below illustrates the arrays that is extracted from a array with column=5:

        [] = read-out electronics
        [==========] = read-out register
        [..........] = serial prescan
        [pppppppppp] = parallel overscan
        [ssssssssss] = serial overscan
        [f#ff#f#f#f] = signal region (FPR) (0 / 1 indicate the region index)
        [tttttttttt] = parallel / serial charge injection region trail

               [ppppppppppppppppppppp]
               [pppppppppppppppppppp ]
          [...][xxxxxxxxxxxxxxxxxxxxx][sss]
          [...][ccccccccccccccccccccc][tst]
        | [...][ccccccccccccccccccccc][sts]    |
        | [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | Direction
        P [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | of
        | [...][ccccccccccccccccccccc][tst]    | clocking
          [...][ccccccccccccccccccccc][sts]    |

        []     [=====================]
               <--------Ser---------

        The extracted array keeps just the trails following all charge injection scans and replaces all other
        values with 0s:

        |                                      |
        |      [cccccccccccccccc][tst]         | Direction
        P      [cccccccccccccccc][tst]         | of
        |      [cccccccccccccccc][tst]         | clocking
               [cccccccccccccccc][tst]         |

        []     [=====================]
               <--------Ser---------
        """
        calibration_images = self.array_2d_list_from(array=array)
        calibration_images = list(
            map(lambda image: image[rows[0] : rows[1], :], calibration_images)
        )

        new_array = np.concatenate(calibration_images, axis=0)

        mask_2d = self.mask_2d_from(mask=array.mask, rows=rows)

        return aa.Array2D(
            values=new_array,
            mask=mask_2d,
            header=array.header,
        )

    def extracted_layout_from(self, layout, new_shape_2d, rows):
        serial_prescan = (
            (0, new_shape_2d[0], self.serial_prescan[2], self.serial_prescan[3])
            if self.serial_prescan is not None
            else None
        )
        serial_overscan = (
            (0, new_shape_2d[0], self.serial_overscan[2], self.serial_overscan[3])
            if self.serial_overscan is not None
            else None
        )

        x0 = self.region_list[0][2]
        x1 = self.region_list[0][3]
        offset = 0

        new_pattern_region_list_ci = []

        for region in self.region_list:
            labelsize = rows[1] - rows[0]
            new_pattern_region_list_ci.append(
                aa.Region2D(region=(offset, offset + labelsize, x0, x1))
            )
            offset += labelsize

        new_layout = deepcopy(layout)
        new_layout.region_list = new_pattern_region_list_ci
        new_layout.serial_prescan = serial_prescan
        new_layout.serial_overscan = serial_overscan

        return new_layout

    def imaging_ci_from(self, dataset: ImagingCI, rows) -> ImagingCI:
        """
        Returnss a function to extract a serial section for given rows
        """

        from autocti.charge_injection.imaging.imaging import ImagingCI

        cosmic_ray_map = (
            dataset.layout.extract.serial_calibration.array_2d_from(
                array=dataset.cosmic_ray_map, rows=rows
            )
            if dataset.cosmic_ray_map is not None
            else None
        )

        if dataset.noise_scaling_map_dict is not None:
            noise_scaling_map_dict = {
                key: dataset.layout.extract.serial_calibration.array_2d_from(
                    array=noise_scaling_map, rows=rows
                )
                for key, noise_scaling_map in dataset.noise_scaling_map_dict.items()
            }

        else:
            noise_scaling_map_dict = None

        image = dataset.layout.extract.serial_calibration.array_2d_from(
            array=dataset.data, rows=rows
        )

        mask = self.mask_2d_from(mask=dataset.mask, rows=rows)

        dataset = ImagingCI(
            data=image,
            noise_map=dataset.layout.extract.serial_calibration.array_2d_from(
                array=dataset.noise_map, rows=rows
            ),
            pre_cti_data=dataset.layout.extract.serial_calibration.array_2d_from(
                array=dataset.pre_cti_data, rows=rows
            ),
            layout=dataset.layout.extract.serial_calibration.extracted_layout_from(
                layout=dataset.layout, new_shape_2d=image.shape, rows=rows
            ),
            cosmic_ray_map=cosmic_ray_map,
            noise_scaling_map_dict=noise_scaling_map_dict,
            fpr_value=dataset.fpr_value,
            settings_dict=dataset.settings_dict,
        )

        return dataset.apply_mask(mask=mask)



import numpy as np
from typing import List, Optional, Tuple

import autoarray as aa


class SettingsExtract:
    def __init__(
        self,
        pixels: Optional[Tuple[int, int]] = None,
        pixels_from_end: Optional[int] = None,
        force_same_row_size: bool = False,
    ):
        """
        Class containing settings used by the `Extract` classes, which perform the extraction of a 2D array into
        another other 2D or 1D arrays.

        Parameters
        ----------
        pixels
            The row or column pixel indexes to extract the region (e.g. FPR / EPER) between. For example, for the
            ``Extract2DParallelFPR`` object the input `pixels=(0, 3)` extracts the 1st, 2nd and 3rd FPR rows). For
           ``Extract2DSerialFPR` objects the input `pixels=(0, 3)` extracts the 1st, 2nd and 3rd FPR columns).
        pixels_from_end
            Alternative row or column pixel index specification, which extracts this number of pixels from the end of
            each region (e.g. FPR / EPER). For example, if each FPR is 100 pixels and `pixels_from_end=10`, the
            last 10 pixels of each FPR (pixels (90, 100)) are extracted.
        force_same_row_size
            If `True`, the returned arrays are forced to have the same number of rows. This is useful for parallel
            EPER regions, where the number of rows in each region is not the same. If `False`, the returned arrays
            have the same number of rows as the region they were extracted from.
        """
        self.pixels = pixels
        self.pixels_from_end = pixels_from_end
        self.force_same_row_size = force_same_row_size

    def region_list_from(self, region_list: List[aa.Region2D]) -> List[aa.Region2D]:
        if self.force_same_row_size:
            row_size = np.min([region.total_rows for region in region_list])

            region_list = [
                aa.Region2D(
                    region=(region.y0, region.y0 + row_size, region.x0, region.x1)
                )
                for region in region_list
            ]

        return region_list



import copy
import numpy as np
from typing import List, Optional, Tuple

import autoarray as aa

from autocti.extract.settings import SettingsExtract


class Extract1D:
    def __init__(
        self,
        shape_1d: Optional[Tuple[int]] = None,
        region_list: Optional[aa.type.Region1DList] = None,
        prescan: Optional[aa.type.Region1DLike] = None,
        overscan: Optional[aa.type.Region1DLike] = None,
    ):
        """
        Abstract class containing methods for extracting regions from a 1D line dataset which contains some sort of
        original signal whose profile before CTI is known (e.g. warm pixel, charge injection).

        This uses the `region_list`, which contains the signal's regions in pixel coordinates (x0, x1).

        Parameters
        ----------
        region_list
            Integer pixel coordinates specifying the corners of signal (x0, x1).
        """

        self.shape_1d = shape_1d

        self.region_list = (
            list(map(aa.Region1D, region_list)) if region_list is not None else None
        )

        if isinstance(prescan, tuple):
            prescan = aa.Region1D(region=prescan)

        if isinstance(overscan, tuple):
            overscan = aa.Region1D(region=overscan)

        self.prescan = prescan
        self.overscan = overscan

    @property
    def total_pixels_min(self) -> int:
        """
        The number of rows between the read-out electronics and the signal closest to them.
        """
        return np.min([region.total_pixels for region in self.region_list])

    @property
    def parallel_rows_between_regions(self):
        return [
            self.region_list[i + 1].x0 - self.region_list[i].x1
            for i in range(len(self.region_list) - 1)
        ]

    @property
    def trail_size_to_array_edge(self):
        return self.shape_1d[0] - np.max([region.x1 for region in self.region_list])

    def region_list_from(self, settings: SettingsExtract) -> List[aa.Region2D]:
        raise NotImplementedError

    def array_1d_list_from(
        self, array: aa.Array1D, settings: SettingsExtract
    ) -> List[aa.Array1D]:
        """
        Extract a specific region from every region on the line dataset and return as a list of 1D arrays.

        For example, this might extract the EPERs trailing every signal.

        The `region_list_from()` of each `Extract1D` class describes the exact extraction performed for each
        extract when this function is called.

        Parameters
        ----------
        array
            The array from which the regions are extracted and put into the returned list of rrays.
        pixels
            The integer range of pixels between which the extraction is performed.
        """

        arr_list = [
            array.native[region.slice]
            for region in self.region_list_from(settings=settings)
        ]

        mask_1d_list = [
            array.mask[region.slice]
            for region in self.region_list_from(settings=settings)
        ]

        return [
            aa.Array1D(values=arr, mask=mask_1d).native
            for arr, mask_1d in zip(arr_list, mask_1d_list)
        ]

    def stacked_array_1d_from(
        self, array: aa.Array1D, settings: SettingsExtract
    ) -> aa.Array1D:
        """
        Extract a region (e.g. the FPR) of every region on the line dataset and stack them by taking their mean.

        This returns the 1D average of the extracted regions (e.g. the FPRs) of all of the regions, which for certain
        calibration datasets one may expect to be similar.

        For fits to a line dataset this function is also used to create images like the stacked 1D residuals,
        which therefore quantify the goodness-of-fit of a CTI model.

        Parameters
        ----------
        array
            The 1D array which contains the line dataset from which the regions (e.g. the FPRs) are extracted and
            stacked.
        pixels
            The pixel index to extract the region between (e.g. `pixels=(0, 3)` extracts the 1st, 2nd and 3rd
            FPR rows)
        """

        arr_list = np.asarray(
            [
                np.array(array.native[region.slice])
                for region in self.region_list_from(settings=settings)
            ]
        )

        mask_1d_list = [
            array.mask[region.slice]
            for region in self.region_list_from(settings=settings)
        ]

        stacked_array_1d = np.mean(
            np.asarray(arr_list), axis=0, where=np.invert(np.asarray(mask_1d_list))
        )

        return aa.Array1D(
            values=np.asarray(stacked_array_1d.data),
            mask=sum(mask_1d_list) == len(mask_1d_list),
        ).native

    def add_to_array(
        self, new_array: aa.Array1D, array: aa.Array1D, settings: SettingsExtract
    ) -> aa.Array1D:
        """
        Extracts the region (e.g. the FPRs) from the line dataset and adds them to a line dataset.

        Parameters
        ----------
        new_array
            The 1D array which the extracted region (e.g. the FPRs) are added to.
        array
            The 1D array which contains the 1D line dataset from which the region (e.g. the FPRs) are extracted.
        pixels
            The row pixel index which determines the region extracted (e.g. `pixels=(0, 3)` will compute the region
            corresponding to the 1st, 2nd and 3rd pixels).
        """
        region_list = self.region_list_from(settings=settings)

        array_1d_list = self.array_1d_list_from(array=array, settings=settings)

        for arr, region in zip(array_1d_list, region_list):
            new_array[region.x0 : region.x1] += arr

        return new_array

    def add_gaussian_noise_to(
        self,
        array: aa.Array1D,
        settings: SettingsExtract,
        noise_sigma: float,
        noise_seed: int = -1,
    ) -> aa.Array2D:
        """
        Adds Gaussian noise of an input sigma value to the regions of the `Extract` object and returns the overall
        input array with this noise added.

        Parameters
        ----------
        array
            The 1D array which contains the charge injection where regions of Gaussian noise is added.
        noise_sigma
            The sigma value (standard deviation) of the Gaussian from which noise values are drann.
        noise_seed
            The seed of the random number generator, used for the random noises maps.
        pixels
            The row pixel index which determines the region of the FPR (e.g. `pixels=(0, 3)` will compute the region
            corresponding to the 1st, 2nd and 3rd FPR rows).
        pixels_from_end
            Alternative row pixex index specification, which extracts this number of pixels from the end of
            each region (e.g. FPR / EPER). For example, if each FPR is 100 pixels and `pixels_from_end=10`, the
            last 10 pixels of each FPR (pixels (90, 100)) are extracted.
        """

        region_list = self.region_list_from(settings=settings)

        array_1d_list = self.array_1d_list_from(array=array, settings=settings)

        array = copy.copy(array.native)

        for arr, region in zip(array_1d_list, region_list):
            array[region.x0 : region.x1] = aa.preprocess.data_with_gaussian_noise_added(
                data=arr, sigma=noise_sigma, seed=noise_seed
            )

        return array



from autocti.aggregator.cti import CTIAgg
from autocti.aggregator.dataset_1d import Dataset1DAgg
from autocti.aggregator.imaging_ci import ImagingCIAgg
from autocti.aggregator.fit_dataset_1d import FitDataset1DAgg
from autocti.aggregator.fit_imaging_ci import FitImagingCIAgg



import logging

from autoarray.structures.header import Header

from autoarray import exc

from autocti.instruments.acs import acs_util

logging.basicConfig()
logger = logging.getLogger()
logger.setLevel("INFO")


class HeaderACS(Header):
    def __init__(
        self,
        header_sci_obj,
        header_hdu_obj,
        quadrant_letter=None,
        hdu=None,
        bias=None,
        bias_serial_prescan_column=None,
    ):
        super().__init__(header_sci_obj=header_sci_obj, header_hdu_obj=header_hdu_obj)

        self.bias = bias
        self.bias_serial_prescan_column = bias_serial_prescan_column
        self.quadrant_letter = quadrant_letter
        self.hdu = hdu

    @property
    def bscale(self):
        return self.header_hdu_obj["BSCALE"]

    @property
    def bzero(self):
        return self.header_hdu_obj["BZERO"]

    @property
    def gain(self):
        return self.header_sci_obj["CCDGAIN"]

    @property
    def calibrated_gain(self):
        if round(self.gain) == 1:
            calibrated_gain = [0.99989998, 0.97210002, 1.01070000, 1.01800000]
        elif round(self.gain) == 2:
            calibrated_gain = [2.002, 1.945, 2.028, 1.994]
        elif round(self.gain) == 4:
            calibrated_gain = [4.011, 3.902, 4.074, 3.996]
        else:
            raise exc.ArrayException(
                "Calibrated gain of ACS file does not round to 1, 2 or 4."
            )

        if self.quadrant_letter == "A":
            return calibrated_gain[0]
        elif self.quadrant_letter == "B":
            return calibrated_gain[1]
        elif self.quadrant_letter == "C":
            return calibrated_gain[2]
        elif self.quadrant_letter == "D":
            return calibrated_gain[3]

    @property
    def original_units(self):
        return self.header_hdu_obj["BUNIT"]

    @property
    def bias_file(self):
        return self.header_sci_obj["BIASFILE"].replace("jref$", "")

    def array_eps_to_counts(self, array_eps):
        return acs_util.array_eps_to_counts(
            array_eps=array_eps, bscale=self.bscale, bzero=self.bzero
        )

    def array_original_to_electrons(self, array, use_calibrated_gain):
        if self.original_units in "COUNTS":
            array = (array * self.bscale) + self.bzero
        elif self.original_units in "CPS":
            array = (array * self.exposure_time * self.bscale) + self.bzero

        if use_calibrated_gain:
            return array * self.calibrated_gain
        else:
            return array * self.gain

    def array_electrons_to_original(self, array, use_calibrated_gain):
        if use_calibrated_gain:
            array /= self.calibrated_gain
        else:
            array /= self.gain

        if self.original_units in "COUNTS":
            return (array - self.bzero) / self.bscale
        elif self.original_units in "CPS":
            return (array - self.bzero) / (self.exposure_time * self.bscale)



from __future__ import annotations
import logging
from typing import TYPE_CHECKING, List, Union

if TYPE_CHECKING:
    from autocti.model.model_util import CTI1D
    from autocti.model.model_util import CTI2D

import autofit as af

from autocti.aggregator.abstract import AbstractAgg

logger = logging.getLogger(__name__)


def _cti_from(fit: af.Fit) -> Union[CTI1D, CTI2D]:
    """
    Returns a list of `CTI` objects from a `PyAutoFit` sqlite database `Fit` object.

    The results of a model-fit can be stored in a sqlite database, including the following attributes of the fit:

    - The model and its best fit parameters (e.g. `model.json`).

    Each individual attribute can be loaded from the database via the `fit.value()` method.

    This method combines all of these attributes and returns a `CTI` object for a given non-linear search sample
    (e.g. the maximum likelihood model). This includes associating adapt images with their respective galaxies.

    If multiple `CTI` objects were fitted simultaneously via analysis summing, the `fit.child_values()` method
    is instead used to load lists of Tracers. This is necessary if each Tracer has different galaxies (e.g. certain
    parameters vary across each dataset and `Analysis` object).

    Parameters
    ----------
    fit
        A PyAutoFit database Fit object containing the generators of the results of model-fits.

    Returns
    -------
    CTI
        The list of CTI models computed via an instance of clocker.
    """

    if len(fit.children) > 0:
        logger.info(
            """
            Using database for a fit with multiple summed Analysis objects.

            CTI objects do not fully support this yet (e.g. adapt images may not be set up correctly)
            so proceed with caution!
            """
        )

    return fit.instance.cti


class CTIAgg(AbstractAgg):
    """
    Interfaces with an `PyAutoFit` aggregator object to create instances of `CTI` objects from the results
    of a model-fit.

    The results of a model-fit can be stored in a sqlite database, including the following attributes of the fit:

    - The model and its best fit parameters (e.g. `model.json`).

    The `aggregator` contains the path to each of these files, and they can be loaded individually. This class
    can load them all at once and create an `CTI` object via the `_cti_from` method.

    This class's methods returns generators which create the instances of the `CTI` objects. This ensures
    that large sets of results can be efficiently loaded from the hard-disk and do not require storing all
    `CTI` instances in the memory at once.

    For example, if the `aggregator` contains 3 model-fits, this class can be used to create a generator which
    creates instances of the corresponding 3 `CTI` objects.

    If multiple `CTI` objects were fitted simultaneously via analysis summing, the `fit.child_values()` method
    is instead used to load lists of CTIs. This is necessary if each CTI has different galaxies (e.g. certain
    parameters vary across each dataset and `Analysis` object).

    This can be done manually, but this object provides a more concise API.

    Parameters
    ----------
    aggregator
        A `PyAutoFit` aggregator object which can load the results of model-fits.
    """

    def object_via_gen_from(self, fit, cti: Union[CTI1D, CTI2D]) -> Union[CTI1D, CTI2D]:
        """
        Returns a generator of `CTI` objects from an input aggregator.

        See `__init__` for a description of how the `CTI` objects are created by this method.

        Parameters
        ----------
        fit
            A `PyAutoFit` `Fit` object which contains the results of a model-fit as an entry in a sqlite database.
        cti
            A CTI model corresponding to a sample of the non-linear search.

        Returns
        -------
        CTI
            A list of `CTI` objects whose parameters are a sample of the non-linear search.
        """
        return _cti_from(fit=fit)
